# Copyright (C) 2005-2010 Osmo Salomaa
#
# This file is part of Gaupol.
#
# Gaupol is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Gaupol is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Gaupol. If not, see <http://www.gnu.org/licenses/>.

"""User interface container and controller for :class:`aeidon.Project`."""

import aeidon
import gaupol
import gtk
import os
import pango
import sys
_ = aeidon.i18n._

__all__ = ("Page",)


class Page(aeidon.Observable):

    """User interface container and controller for :class:`aeidon.Project`.

    :ivar edit_mode: :attr:`aeidon.modes` item corresponding to editing mode
    :ivar project: The associated :class:`aeidon.Project` instance
    :ivar tab_label: :class:`gtk.Label` contained in :attr:`tab_widget`
    :ivar tab_widget: Widget that can be placed in a notebook tab
    :ivar untitle: Title used if the :attr:`project.main_file` is unsaved
    :ivar view: The associated :class:`gaupol.View` instance

    Signals and their arguments for callback functions:
     * ``close-request``: page
     * ``view-created``: page view

    This class represents one page in a notebook of user interfaces for
    projects. The view is updated automatically when project data changes.
    """

    __metaclass__ = aeidon.Contractual
    signals = ("close-request", "view-created")

    def __init__(self, count=0):
        """Initialize a :class:`Page` object."""
        aeidon.Observable.__init__(self)
        self.edit_mode = gaupol.conf.editor.mode
        self.project = None
        self.tab_label = None
        self.tab_widget = None
        self.untitle = _("Untitled %d") % count
        self.view = gaupol.View(self.edit_mode)
        self._init_project()
        self._init_widgets()
        self._init_signal_handlers()
        self.update_tab_label()
        self.emit("view-created", self.view)

    def _assert_store(self, fields=None):
        """Assert that store's data matches project's."""
        if fields is None:
            fields = list(gaupol.fields)
            fields.remove(gaupol.fields.NUMBER)
        mode = self.edit_mode
        store = self.view.get_model()
        assert len(store) == len(self.project.subtitles)
        for i, subtitle in enumerate(self.project.subtitles):
            if gaupol.fields.START in fields:
                assert store[i][1] == subtitle.get_start(mode)
            if gaupol.fields.END in fields:
                assert store[i][2] == subtitle.get_end(mode)
            if gaupol.fields.DURATION in fields:
                if mode == aeidon.modes.TIME:
                    assert store[i][3] == subtitle.duration_seconds
                if mode == aeidon.modes.FRAME:
                    assert store[i][3] == subtitle.duration_frame
            if gaupol.fields.MAIN_TEXT in fields:
                assert store[i][4] == subtitle.main_text
            if gaupol.fields.TRAN_TEXT in fields:
                assert store[i][5] == subtitle.tran_text

    def _get_subtitle_value(self, row, field):
        """Return value of subtitle data for `row` and `field`."""
        mode = self.edit_mode
        subtitle = self.project.subtitles[row]
        if field == gaupol.fields.START:
            return subtitle.get_start(mode)
        if field == gaupol.fields.END:
            return subtitle.get_end(mode)
        if field == gaupol.fields.DURATION:
            if mode == aeidon.modes.TIME:
                return subtitle.duration_seconds
            if mode == aeidon.modes.FRAME:
                return subtitle.duration_frame
            raise ValueError("Invalid mode: %s" % repr(mode))
        if field == gaupol.fields.MAIN_TEXT:
            return subtitle.main_text
        if field == gaupol.fields.TRAN_TEXT:
            return subtitle.tran_text
        raise ValueError("Invalid field: %s" % repr(field))

    def _get_tab_close_button(self):
        """Initialize and return a tab close button."""
        button = gtk.Button()
        button.set_name("gaupol-tab-close-button")
        image = gtk.image_new_from_stock(gtk.STOCK_CLOSE, gtk.ICON_SIZE_MENU)
        button.add(image)
        button.set_relief(gtk.RELIEF_NONE)
        button.set_focus_on_click(False)
        width, height = image.size_request()
        padding = (6 if sys.platform == "win32" else 2)
        button.set_size_request(width + padding, height + padding)
        request_close = lambda x, self: self.emit("close-request")
        button.connect("clicked", request_close, self)
        button.set_tooltip_text(_("Close project"))
        return button

    def _init_project(self):
        """Initialize :class:`aeidon.Project` with proper properties."""
        limit = gaupol.conf.editor.use_undo_limit
        levels = gaupol.conf.editor.undo_limit
        undo_limit = (levels if limit else None)
        self.project = aeidon.Project(gaupol.conf.editor.framerate,
                                      undo_limit)

    def _init_signal_handlers(self):
        """Initialize signal handlers."""
        self._init_signal_handlers_for_data()
        self._init_signal_handlers_for_tab_label()
        self._init_signal_handlers_for_undo()

    def _init_signal_handlers_for_data(self):
        """Initialize signal handlers for project data updates."""
        aeidon.util.connect(self, "project", "main-file-opened")
        aeidon.util.connect(self, "project", "main-texts-changed")
        aeidon.util.connect(self, "project", "positions-changed")
        aeidon.util.connect(self, "project", "subtitles-changed")
        aeidon.util.connect(self, "project", "subtitles-inserted")
        aeidon.util.connect(self, "project", "subtitles-removed")
        aeidon.util.connect(self, "project", "translation-file-opened")
        aeidon.util.connect(self, "project", "translation-texts-changed")

    def _init_signal_handlers_for_tab_label(self):
        """Initialize signal handlers for tab label updates."""
        aeidon.util.connect(self, "tab_label", "query-tooltip")
        update_label = lambda *args: args[-1].update_tab_label()
        self.project.connect("main-file-opened", update_label, self)
        self.project.connect("main-file-saved", update_label, self)
        self.project.connect("notify::main_changed", update_label, self)
        self.project.connect("notify::tran_changed", update_label, self)

    def _init_signal_handlers_for_undo(self):
        """Initialize signal handlers for undo level updates."""
        connect = gaupol.conf.editor.connect
        update_levels = lambda *args: args[-1]._update_undo_limit()
        connect("notify::use_undo_limit", update_levels, self)
        connect("notify::undo_limit", update_levels, self)

    def _init_widgets(self):
        """Initialize widgets to use in a notebook tab."""
        self.tab_label = gtk.Label()
        self.tab_label.props.xalign = 0
        self.tab_label.set_ellipsize(pango.ELLIPSIZE_END)
        self.tab_label.set_max_width_chars(24)
        self.tab_label.set_tooltip_text(self.untitle)
        button = self._get_tab_close_button()
        self.tab_widget = gtk.HBox(False, 4)
        self.tab_widget.pack_start(self.tab_label, True, True, 0)
        self.tab_widget.pack_start(button, False, False, 0)
        self.tab_widget.set_data("button", button)
        self.tab_widget.show_all()

    def _on_project_main_file_opened(self, *args):
        """Reload the entire view."""
        self.reload_view_all()
        gaupol.util.iterate_main()

    def _on_project_main_texts_changed(self, project, rows):
        """Reload and select main texts in rows."""
        if not rows: return
        fields = (gaupol.fields.MAIN_TEXT,)
        self.reload_view(rows, fields)
        if self.view.get_focus()[0] not in rows:
            col = self.view.columns.MAIN_TEXT
            self.view.set_focus(rows[0], col)
        self.view.select_rows(rows)
        gaupol.util.iterate_main()

    def _on_project_positions_changed(self, project, rows):
        """Reload and select positions in rows."""
        if not rows: return
        enum = gaupol.fields
        fields = (enum.START, enum.END, enum.DURATION)
        self.reload_view(rows, fields)
        if self.view.get_focus()[0] not in rows:
            self.view.set_focus(rows[0])
        self.view.select_rows(rows)
        gaupol.util.iterate_main()

    def _on_project_subtitles_changed(self, project, rows):
        """Reload and select subtitles in rows."""
        if not rows: return
        fields = [x for x in gaupol.fields]
        fields.remove(gaupol.fields.NUMBER)
        self.reload_view(rows, fields)
        if self.view.get_focus()[0] not in rows:
            self.view.set_focus(rows[0])
        self.view.select_rows(rows)
        gaupol.util.iterate_main()

    def _on_project_subtitles_inserted_ensure(self, value, project, rows):
        self._assert_store()

    def _on_project_subtitles_inserted(self, project, rows):
        """Insert rows to the view and select them."""
        if not rows: return
        mode = self.edit_mode
        store = self.view.get_model()
        for row in sorted(rows):
            subtitle = self.project.subtitles[row]
            store.insert(row)
            store[row][0] = row + 1
            store[row][1] = subtitle.get_start(mode)
            store[row][2] = subtitle.get_end(mode)
            if mode == aeidon.modes.TIME:
                store[row][3] = subtitle.duration_seconds
            if mode == aeidon.modes.FRAME:
                store[row][3] = subtitle.duration_frame
            store[row][4] = subtitle.main_text
            store[row][5] = subtitle.tran_text
        self.view.set_focus(rows[0])
        self.view.select_rows(rows)
        gaupol.util.iterate_main()

    def _on_project_subtitles_removed_ensure(self, value, project, rows):
        self._assert_store()

    def _on_project_subtitles_removed(self, project, rows):
        """Remove rows from the view."""
        if not rows: return
        store = self.view.get_model()
        if len(rows) > 50:
            # Unset and later reset the model if removing a large amount of
            # rows, because a large batch of separate live updates directly
            # made to the view are slow, 50 being a rough empirical border.
            self.view.set_model(None)
        for row in reversed(sorted(rows)):
            store.remove(store.get_iter(row))
        if len(rows) > 50:
            self.view.set_model(store)
        if self.project.subtitles:
            row = min(rows[0], len(self.project.subtitles) - 1)
            col = self.view.get_focus()[1]
            self.view.set_focus(row, col)
        gaupol.util.iterate_main()

    def _on_project_translation_file_opened(self, *args):
        """Reload the entire view."""
        self.reload_view_all()
        gaupol.util.iterate_main()

    def _on_project_translation_texts_changed(self, project, rows):
        """Reload and select translation texts in rows."""
        if not rows: return
        fields = (gaupol.fields.TRAN_TEXT,)
        self.reload_view(rows, fields)
        if self.view.get_focus()[0] not in rows:
            col = self.view.columns.TRAN_TEXT
            self.view.set_focus(rows[0], col)
        self.view.select_rows(rows)
        gaupol.util.iterate_main()

    def _on_tab_label_query_tooltip(self, label, x, y, keyboard, tooltip):
        """Update the text in the tab tooltip."""
        if self.project.main_file is None: return
        encoding = self.project.main_file.encoding
        encoding = aeidon.encodings.code_to_long_name(encoding)
        tooltip.set_markup("%s\n\n%s\n%s\n%s" % (
            _("<b>Path:</b> %s") % self.project.main_file.path,
            _("<b>Format:</b> %s") % self.project.main_file.format.label,
            _("<b>Character encoding:</b> %s") % encoding,
            _("<b>Newlines:</b> %s") % self.project.main_file.newline.label))

        return True # to show the tooltip.

    def _update_undo_limit(self):
        """Update project's undo level count to match global configuration."""
        self.project.undo_limit = (gaupol.conf.editor.undo_limit if
                                   gaupol.conf.editor.use_undo_limit else None)

    def document_to_text_column(self, doc):
        """Translate document enumeration to view's column enumeration."""
        if doc == aeidon.documents.MAIN:
            return self.view.columns.MAIN_TEXT
        if doc == aeidon.documents.TRAN:
            return self.view.columns.TRAN_TEXT
        raise ValueError("Invalid document: %s" % repr(doc))

    def get_main_basename(self):
        """Return basename of the main document."""
        if self.project.main_file is not None:
            return os.path.basename(self.project.main_file.path)
        return self.untitle

    def get_translation_basename(self):
        """Return basename of the translation document."""
        if self.project.tran_file is not None:
            return os.path.basename(self.project.tran_file.path)
        basename = self.get_main_basename()
        if self.project.main_file is not None:
            extension = self.project.main_file.format.extension
            if basename.endswith(extension):
                basename = basename[:-len(extension)]
        return _("%s translation") % basename

    def reload_view_ensure(self, value, rows, fields):
        self._assert_store(fields)

    def reload_view(self, rows, fields):
        """Reload the view in `rows` and `fields`."""
        store = self.view.get_model()
        for row, field in ((x, y) for x in rows for y in fields):
            value = self._get_subtitle_value(row, field)
            store[row][field] = value

    def reload_view_all_ensure(self, value):
        self._assert_store()

    def reload_view_all(self):
        """Clear and repopulate the entire view."""
        store = self.view.get_model()
        self.view.set_model(None)
        store.clear()
        mode = self.edit_mode
        for i, subtitle in enumerate(self.project.subtitles):
            store.insert(i)
            store[i][0] = i + 1
            store[i][1] = subtitle.get_start(mode)
            store[i][2] = subtitle.get_end(mode)
            if mode == aeidon.modes.TIME:
                store[i][3] = subtitle.duration_seconds
            if mode == aeidon.modes.FRAME:
                store[i][3] = subtitle.duration_frame
            store[i][4] = subtitle.main_text
            store[i][5] = subtitle.tran_text
        self.view.set_model(store)

    def text_column_to_document(self, col):
        """Translate view's column enumeration to document enumeration."""
        if col == self.view.columns.MAIN_TEXT:
            return aeidon.documents.MAIN
        if col == self.view.columns.TRAN_TEXT:
            return aeidon.documents.TRAN
        raise ValueError("Invalid column: %s" % repr(col))

    def update_tab_label(self):
        """Update the notebook tab label and return title."""
        title = self.get_main_basename()
        if self.project.main_changed or self.project.tran_changed:
            title = "*%s" % title
        self.tab_label.set_text(title)
        return title



# Copyright (C) 2006-2010 Osmo Salomaa
#
# This file is part of Gaupol.
#
# Gaupol is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Gaupol is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Gaupol. If not, see <http://www.gnu.org/licenses/>.

"""Observable dictionary with attribute access to keys."""

import aeidon

__all__ = ("AttributeDictionary",)


class AttributeDictionary(aeidon.Observable):

    """Observable dictionary with attribute access to keys.

    :class:`AttributeDictionary` is initialized from a root dictionary, which
    is kept in sync with attribute values. This allows convenient attribute
    access to dictionary keys and notifications of changes via the
    :class:`aeidon.Observable` interface.
    """

    def __init__(self, root):
        """Initialize an AttributeDictionary object.

        All subdictionaries of `root` are initialized as instances of
        :class:`AttributeDictionary` as well.
        """
        aeidon.Observable.__init__(self)
        self._root = root
        self.update(root)

    def _on_notify(self, obj, value, name):
        """Synchronize changed attribute value with root dictionary."""
        self._root[name] = value

    def add_attribute_ensure(self, rvalue, name, value):
        assert self._root[name] == value
        if not isinstance(value, dict):
            assert getattr(self, name) == value

    def add_attribute(self, name, value):
        """Add instance attribute and corresponding root dictionary key."""
        self._root[name] = value
        # In the case of dictionaries, set the original dictionary to the root
        # dictionary, but instantiate an AttributeDictionary for use as the
        # corresponding attribute.
        if isinstance(value, dict):
            value = AttributeDictionary(value)
        setattr(self, name, value)
        self.connect("notify::%s" % name, self._on_notify, name)

    def extend(self, root):
        """Add new values from another root dictionary."""
        for name, value in root.iteritems():
            if not hasattr(self, name):
                self.add_attribute(name, value)
            else: # Extend current value.
                if isinstance(value, dict):
                    getattr(self, name).extend(value)

    def remove_attribute_ensure(self, rvalue, name):
        assert not hasattr(self, name)
        assert not name in self._root

    def remove_attribute(self, name):
        """Remove instance attribute and corresponding root dictionary key."""
        self.disconnect("notify::%s" % name, self._on_notify)
        delattr(self, name)
        if name in self._root:
            del self._root[name]

    def update(self, root):
        """Update values from another root dictionary."""
        for name, value in root.iteritems():
            if not hasattr(self, name):
                self.add_attribute(name, value)
            else: # Update current value.
                if isinstance(value, dict):
                    getattr(self, name).update(value)
                else: setattr(self, name, value)



# Copyright (C) 2005-2008 Osmo Salomaa
#
# This file is part of Gaupol.
#
# Gaupol is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Gaupol is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Gaupol. If not, see <http://www.gnu.org/licenses/>.

import aeidon
import gaupol


class TestAction(gaupol.TestCase):

    def setup_method(self, method):
        self.action = gaupol.Action("test")
        self.action.accelerator = "X"
        self.action.widgets = ("notebook",)
        self.application = gaupol.Application()
        self.application._on_test_activate = lambda *args: None
        self.action.finalize(self.application)

    def test_set_sensitive(self):
        self.action.set_sensitive(False)
        self.action.set_sensitive(True)

    def test_update_sensitivity(self):
        self.action.update_sensitivity(self.application, None)


class TestMenuAction(gaupol.TestCase):

    def setup_method(self, method):
        self.action = gaupol.MenuAction("test")
        self.action.widgets = ("notebook",)
        self.application = gaupol.Application()
        self.application._on_test_activate = lambda *args: None
        self.action.finalize(self.application)


class TestRecentAction(gaupol.TestCase):

    @aeidon.deco.monkey_patch(gaupol.RecentAction, "group")
    def setup_method(self, method):
        gaupol.RecentAction.group = "gaupol-main"
        self.action = gaupol.RecentAction("test")
        self.action.widgets = ("notebook",)
        self.application = gaupol.Application()
        self.application._on_test_item_activated = lambda *args: None
        self.action.finalize(self.application)


class TestToggleAction(gaupol.TestCase):

    def setup_method(self, method):
        self.action = gaupol.ToggleAction("test")
        self.action.widgets = ("notebook",)
        self.application = gaupol.Application()
        self.application._on_test_toggled = lambda *args: None
        self.action.finalize(self.application)


class TestTopMenuAction(gaupol.TestCase):

    def setup_method(self, method):
        self.action = gaupol.TopMenuAction("test")
        self.action.widgets = ("notebook",)
        self.application = gaupol.Application()
        self.action.finalize(self.application)


class TestRadioAction(gaupol.TestCase):

    def setup_method(self, method):
        self.action = gaupol.RadioAction("test")
        self.action.widgets = ("notebook",)
        self.action.group = self.action.__class__.__name__
        self.application = gaupol.Application()
        self.application._on_test_changed = lambda *args: None
        self.action.finalize(self.application)



# Copyright (C) 2005-2010 Osmo Salomaa
#
# This file is part of Gaupol.
#
# Gaupol is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Gaupol is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Gaupol. If not, see <http://www.gnu.org/licenses/>.

import aeidon
import gaupol
import gtk
import random


class TestView(gaupol.TestCase):

    def run__view_frame(self):
        self.setup_time()
        window = gtk.Window()
        window.connect("delete-event", gtk.main_quit)
        window.set_position(gtk.WIN_POS_CENTER)
        window.set_default_size(200, 200)
        window.add(self.view)
        window.show_all()
        gtk.main()

    def run__view_time(self):
        self.setup_frame()
        window = gtk.Window()
        window.connect("delete-event", gtk.main_quit)
        window.set_position(gtk.WIN_POS_CENTER)
        window.set_default_size(200, 200)
        window.add(self.view)
        window.show_all()
        gtk.main()

    def setup_frame(self):
        # pylint: disable=W0201
        self.view = gaupol.View(aeidon.modes.FRAME)
        store = self.view.get_model()
        self.project = self.new_project()
        for i, subtitle in enumerate(self.project.subtitles):
            store.append((i,
                          subtitle.start_frame,
                          subtitle.end_frame,
                          subtitle.duration_frame,
                          subtitle.main_text,
                          subtitle.tran_text,))

    def setup_method(self, method):
        random.choice((self.setup_frame, self.setup_time))()
        self.conf = gaupol.conf.editor

    def setup_time(self):
        # pylint: disable=W0201
        self.view = gaupol.View(aeidon.modes.TIME)
        store = self.view.get_model()
        self.project = self.new_project()
        for i, subtitle in enumerate(self.project.subtitles):
            store.append((i,
                          subtitle.start_time,
                          subtitle.end_time,
                          subtitle.duration_seconds,
                          subtitle.main_text,
                          subtitle.tran_text,))

    def test__on_conf_editor_notify_custom_font(self):
        self.conf.use_custom_font = True
        self.conf.custom_font = "Serif 12"

    def test__on_conf_editor_notify_length_unit(self):
        for unit in gaupol.length_units:
            self.conf.length_unit = unit

    def test__on_conf_editor_notify_show_lengths_cell(self):
        self.conf.show_lengths_cell = True
        self.conf.show_lengths_cell = False

    def test__on_conf_editor_notify_use_custom_font(self):
        self.conf.custom_font = "Serif 12"
        self.conf.use_custom_font = True

    def test__on_columns_changed(self):
        for column in self.view.get_columns():
            column.set_visible(False)
        for column in self.view.get_columns():
            column.set_visible(True)

    def test__on_cursor_changed(self):
        self.view.emit("cursor-changed")
        self.view.emit("cursor-changed")

    def test_get_focus(self):
        get_focus = self.view.get_focus
        self.view.set_focus(2, None)
        assert get_focus() == (2, None)
        self.view.set_focus(1, 3)
        assert get_focus() == (1, 3)

    def test_get_selected_rows(self):
        get_selected_rows = self.view.get_selected_rows
        self.view.select_rows(())
        assert get_selected_rows() == ()
        self.view.select_rows((1, 2))
        assert get_selected_rows() == (1, 2)

    def test_is_position_column(self):
        is_position = self.view.is_position_column
        assert is_position(self.view.columns.START)
        assert is_position(self.view.columns.END)
        assert is_position(self.view.columns.DURATION)
        assert not is_position(self.view.columns.NUMBER)
        assert not is_position(self.view.columns.MAIN_TEXT)
        assert not is_position(self.view.columns.TRAN_TEXT)

    def test_is_text_column(self):
        is_text = self.view.is_text_column
        assert is_text(self.view.columns.MAIN_TEXT)
        assert is_text(self.view.columns.TRAN_TEXT)
        assert not is_text(self.view.columns.NUMBER)
        assert not is_text(self.view.columns.START)
        assert not is_text(self.view.columns.END)
        assert not is_text(self.view.columns.DURATION)

    def test_scroll_to_row(self):
        self.view.scroll_to_row(1)
        self.view.scroll_to_row(2)

    def test_select_rows(self):
        get_selected_rows = self.view.get_selected_rows
        self.view.select_rows(())
        assert get_selected_rows() == ()
        self.view.select_rows((1, 2))
        assert get_selected_rows() == (1, 2)

    def test_set_focus(self):
        get_focus = self.view.get_focus
        self.view.set_focus(2, None)
        assert get_focus() == (2, None)
        self.view.set_focus(1, 3)
        assert get_focus() == (1, 3)
        self.view.set_focus(-1, 4)
        n = len(self.project.subtitles)
        assert get_focus() == (n - 1, 4)

    def test_update_headers(self):
        self.view.update_headers()
        self.view.set_focus(1, 3)
        self.view.update_headers()
        self.view.set_focus(2, 4)
        self.view.update_headers()



# Copyright (C) 2008,2010 Osmo Salomaa
#
# This file is part of Gaupol.
#
# Gaupol is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Gaupol is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Gaupol. If not, see <http://www.gnu.org/licenses/>.

"""A side pane that can be used by other extensions."""

import aeidon
import gaupol
import gtk
import os
_ = aeidon.i18n._


class SidePane(aeidon.Observable):

    """A side pane object for the application window.

    Side pane is installed as an attribute of :class:`gaupol.Application` and
    thus accessible to extensions as ``application.side_pane``.

    Signals and their arguments for callback functions:
     * ``close-button-clicked``: side_pane
     * ``page-switched``: side_pane, new_page
    """

    # Parts of the code dealing with the header's toggle button and the menu
    # associated with it have been adapted from Nautilus, file
    # 'nautilus-side-pane.c' with the following copyright statement and info.
    # Copyright (C) 2002 Ximian Inc.
    # Author: Dave Camp <dave@ximian.com>

    signals = ("close-button-clicked", "page-switched")

    def __init__(self, application):
        """Initialize a :class:`SidePane` object."""
        aeidon.Observable.__init__(self)
        self._conf = gaupol.conf.extensions.side_pane
        self._focus_handler_id = None
        self._has_focus = False
        self._label = gtk.Label("(Empty)")
        self._notebook = gtk.Notebook()
        self._paned = gtk.HPaned()
        self._toggle_button = gtk.ToggleButton()
        self.application = application
        self._init_gui()
        self._init_signal_handlers()

    def _init_gui(self):
        """Initialize all widgets."""
        side_vbox = gtk.VBox(False, 0)
        self._init_paned(side_vbox)
        self._init_header(side_vbox)
        self._init_notebook(side_vbox)
        self._paned.show_all()
        child = self._paned.get_child1()
        child.props.visible = self._conf.visible

    def _init_header(self, side_vbox):
        """Initialize the side pane button header."""
        header_hbox = gtk.HBox(False, 12)
        header_hbox.set_border_width(1)
        self._toggle_button.set_relief(gtk.RELIEF_NONE)
        self._toggle_button.set_focus_on_click(False)
        toggle_button_hbox = gtk.HBox(False, 6)
        toggle_button_hbox.pack_start(self._label)
        arrow = gtk.Arrow(gtk.ARROW_DOWN, gtk.SHADOW_NONE)
        toggle_button_hbox.pack_start(arrow)
        self._toggle_button.add(toggle_button_hbox)
        callback = self._on_header_toggle_button_button_press_event
        self._toggle_button.connect("button-press-event", callback)
        callback = self._on_header_toggle_button_key_press_event
        self._toggle_button.connect("key-press-event", callback)
        header_hbox.pack_start(self._toggle_button, False, False)
        header_hbox.pack_start(gtk.Alignment(), True, True)
        close_button = gtk.Button()
        image = gtk.image_new_from_stock(gtk.STOCK_CLOSE, gtk.ICON_SIZE_MENU)
        close_button.add(image)
        close_button.set_relief(gtk.RELIEF_NONE)
        close_button.set_focus_on_click(False)
        callback = self._on_header_close_button_clicked
        close_button.connect("clicked", callback)
        header_hbox.pack_start(close_button, False, False)
        side_vbox.pack_start(header_hbox, False, False)

    def _init_notebook(self, side_vbox):
        """Initialize the side pane notebook."""
        self._notebook.set_show_border(False)
        self._notebook.set_show_tabs(False)
        side_vbox.pack_start(self._notebook, True, True)

    def _init_paned(self, side_vbox):
        """Initialize the horizontal pane container."""
        main_vbox = self.application.window.get_children()[0]
        main_notebook = main_vbox.get_children()[2]
        self._paned.pack1(side_vbox, False, False)
        main_notebook_vbox = gtk.VBox()
        main_notebook.reparent(main_notebook_vbox)
        self._paned.pack2(main_notebook_vbox, True, False)
        main_vbox.pack_start(self._paned)
        main_vbox.reorder_child(self._paned, 2)
        self._paned.set_position(self._conf.width)

    def _init_signal_handlers(self):
        """Initialize signal handlers."""
        self._focus_handler_id = self.application.window.connect(
            "set-focus", self._on_application_window_set_focus)

    def _on_application_window_set_focus(self, window, widget):
        """Disable unsafe UI manager actions.

        Disabling unsafe UI manager actions allows the side pane to contain any
        widgets that can have input focus and can have their own keybindings,
        without being in conflict with the keybindings of UI manager actions.
        """
        if widget is None: return
        action_group = self.application.get_action_group("main-unsafe")
        in_side_pane = widget.is_ancestor(self._paned.get_child1())
        if self._has_focus or in_side_pane:
            action_group.set_sensitive(not in_side_pane)
        self._has_focus = in_side_pane

    def _on_header_close_button_clicked(self, button):
        """Hide the side pane from the main window."""
        self.hide()
        self.emit("close-button-clicked")

    def _on_header_menu_deactivate(self, menu):
        """Set the header toggle button back to inactive."""
        self._toggle_button.set_active(False)
        menu.destroy()

    def _on_header_menu_item_activate(self, menu_item):
        """Set the currently active page to activated `menu_item`."""
        child = menu_item.get_data("child")
        self.set_current_page(child)
        self._toggle_button.set_active(False)
        parent = menu_item.get_parent()
        while not isinstance(parent, gtk.Menu):
            parent = parent.get_parent()
        parent.destroy()
        self.emit("page-switched", child)

    def _on_header_toggle_button_button_press_event(self, button, event):
        """Show a menu listing all side pane pages."""
        if event.button != 1: return False
        self._show_header_menu(event)
        return True

    def _on_header_toggle_button_key_press_event(self, key, event):
        """Show a menu listing all side pane pages."""
        spaces = (gtk.keysyms.space, gtk.keysyms.KP_Space)
        enters = (gtk.keysyms.Return, gtk.keysyms.KP_Enter)
        if not event.keyval in (spaces + enters): return False
        self._show_header_menu(event)
        return True

    def _position_header_menu(self, menu):
        """Return coordinates for ``menu.popup`` below the toggle button."""
        # pylint: disable=E1101
        x, y, = self._toggle_button.window.get_origin()
        allocation = self._toggle_button.get_allocation()
        x += allocation.x
        y += allocation.y + allocation.height
        return x, y, True

    def _show_header_menu(self, event):
        """Show a menu listing all side pane pages."""
        menu = gtk.Menu()
        for i in range(self._notebook.get_n_pages()):
            child = self._notebook.get_nth_page(i)
            title = self._notebook.get_tab_label_text(child)
            menu_item = gtk.MenuItem(title, False)
            menu_item.set_data("child", child)
            menu_item.connect("activate", self._on_header_menu_item_activate)
            menu.append(menu_item)
        menu.connect("deactivate", self._on_header_menu_deactivate)
        menu.show_all()
        self._toggle_button.set_active(True)
        menu.popup(None, None, self._position_header_menu, 1, event.time)

    def add_page(self, child, name, title):
        """Add `child` as a page to the side pane.

        `name` should be a unique string and is used internally for saving the
        active page name to the configuration file. Using the extension's
        module name is a good idea since that should be unique anyway, or if
        having multiple side pane pages per extension, then a name prefixed
        with the module name. `title` is a string shown to the user.
        """
        child.set_data("side_pane_extension_name", name)
        self._notebook.append_page(child)
        self._notebook.set_tab_label_text(child, title)
        if self._notebook.get_n_pages() == 1:
            self.set_current_page(child)
        elif name == self._conf.page:
            self.set_current_page(child)

    def get_current_page(self):
        """Return the child widget of the currently active page or ``None``."""
        if self._notebook.get_n_pages() == 0: return None
        page_num = self._notebook.get_current_page()
        return self._notebook.get_nth_page(page_num)

    def hide(self):
        """Hide the side pane from the application window."""
        self._paned.get_child1().hide()

    def remove(self):
        """Remove the entire side pane from the application window.

        Use :meth:`hide` unless you really know that you really want to remove
        the side pane. Usually this is used only once the side pane extension
        is deactivated, which is when all extensions using it are deactivated.
        """
        self.application.window.disconnect(self._focus_handler_id)
        self._conf.width = self._paned.get_position()
        child = self.get_current_page()
        if child is not None:
            self._conf.page = child.get_data("side_pane_extension_name")
        child = self._paned.get_child1()
        self._conf.visible = child.props.visible
        main_vbox = self.application.window.get_children()[0]
        main_notebook = self._paned.get_child2().get_children()[0]
        main_vbox.remove(self._paned)
        main_notebook.reparent(main_vbox)
        main_vbox.reorder_child(main_notebook, 2)
        self._paned.destroy()

    def remove_page(self, child):
        """Remove `child` page from the side pane."""
        page_num = self._notebook.page_num(child)
        self._notebook.remove_page(page_num)

    def set_current_page(self, child):
        """Set the currently active page to `child`."""
        page_num = self._notebook.page_num(child)
        self._notebook.set_current_page(page_num)
        title = self._notebook.get_tab_label_text(child)
        self._label.set_text(title)
        self._conf.page = child.get_data("side_pane_extension_name")

    def show(self):
        """Show the side pane in the application window."""
        self._paned.get_child1().show()


class SidePaneExtension(gaupol.Extension):

    """A side pane that can be used by other extensions."""

    def __init__(self):
        """Initialize a :class:`SidePaneExtension` object."""
        self._action_group = None
        self._conf = None
        self._uim_id = None
        self.application = None

    def _on_side_pane_close_button_clicked(self, side_pane):
        """Update the state of the corresponding menu item."""
        path = "/ui/menubar/view/1/side_pane"
        action = self.application.uim.get_action(path)
        action.set_active(False)

    def _on_toggle_side_pane_toggled(self, *args):
        """Show or hide the side pane."""
        if not self._conf.visible:
            self.application.side_pane.show()
        else: # Is visible.
            self.application.side_pane.hide()
        self._conf.visible = not self._conf.visible

    def setup(self, application):
        """Setup extension for use with `application`."""
        gaupol.conf.register_extension("side_pane",
                                       {"width": 200,
                                        "page": "",
                                        "visible": True})

        self._conf = gaupol.conf.extensions.side_pane
        application.side_pane = SidePane(application)
        self._action_group = gtk.ActionGroup("side-pane")
        self._action_group.add_toggle_actions((
            ("toggle_side_pane", None, _("Si_de Pane"),
             None, _("Show or hide the side pane"),
             self._on_toggle_side_pane_toggled, self._conf.visible),))

        application.uim.insert_action_group(self._action_group, -1)
        directory = os.path.abspath(os.path.dirname(__file__))
        ui_file_path = os.path.join(directory, "side-pane.ui.xml")
        self._uim_id = application.uim.add_ui_from_file(ui_file_path)
        callback = self._on_side_pane_close_button_clicked
        application.side_pane.connect("close-button-clicked", callback)
        application.uim.ensure_update()
        application.set_menu_notify_events("side-pane")
        self.application = application

    def teardown(self, application):
        """End use of extension with `application`."""
        application.side_pane.remove()
        del application.side_pane
        self.application.uim.remove_ui(self._uim_id)
        self.application.uim.remove_action_group(self._action_group)
        self.application.uim.ensure_update()



# Copyright (C) 2005-2009 Osmo Salomaa
#
# This file is part of Gaupol.
#
# Gaupol is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Gaupol is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Gaupol. If not, see <http://www.gnu.org/licenses/>.

"""Miscellaneous helper methods."""

import aeidon
import os


class UtilityAgent(aeidon.Delegate):

    """Miscellaneous helper methods."""

    __metaclass__ = aeidon.Contractual

    @aeidon.deco.export
    def get_all_indices(self):
        """Return a list of all indices of subtitles."""
        return range(len(self.subtitles))

    @aeidon.deco.export
    def get_changed(self, doc):
        """Return the changed value corresponding to `doc`."""
        if doc == aeidon.documents.MAIN:
            return self.main_changed
        if doc == aeidon.documents.TRAN:
            return self.tran_changed
        raise ValueError("Invalid document: %s" % repr(doc))

    @aeidon.deco.export
    def get_file(self, doc):
        """Return the file corresponding to `doc`."""
        if doc == aeidon.documents.MAIN:
            return self.main_file
        if doc == aeidon.documents.TRAN:
            return self.tran_file
        raise ValueError("Invalid document: %s" % repr(doc))

    @aeidon.deco.export
    def get_format(self, doc):
        """Return format of the file corresponding to `doc`.

        For translation file that is ``None``, return format of main file.
        If main file is ``None``, return ``None``.
        """
        if doc == aeidon.documents.MAIN:
            if self.main_file is not None:
                return self.main_file.format
            return None
        if doc == aeidon.documents.TRAN:
            if self.tran_file is not None:
                return self.tran_file.format
            return self.get_format(aeidon.documents.MAIN)
        raise ValueError("Invalid document: %s" % repr(doc))

    @aeidon.deco.export
    def get_liner(self, doc):
        """Return a new :class:`aeidon.Liner` instance."""
        re_tag = self.get_markup_tag_regex(doc)
        clean_func = self.get_markup_clean_func(doc)
        return aeidon.Liner(re_tag, clean_func)

    @aeidon.deco.export
    def get_markup(self, doc):
        """Return `doc`'s markup instance or ``None``."""
        format = self.get_format(doc)
        if format is None: return None
        return aeidon.tags.new(format)

    @aeidon.deco.export
    def get_markup_clean_func(self, doc):
        """Return the function to clean markup or ``None``."""
        format = self.get_format(doc)
        if format is None: return None
        return aeidon.tags.new(format).clean

    @aeidon.deco.export
    def get_markup_tag_regex(self, doc):
        """Return the regular expression for a markup tag or ``None``."""
        format = self.get_format(doc)
        if format is None: return None
        return aeidon.tags.new(format).tag

    @aeidon.deco.export
    def get_mode(self):
        """Return the mode of the main file or default."""
        if self.main_file is not None:
            return self.main_file.mode
        return aeidon.modes.TIME

    @aeidon.deco.export
    def get_parser(self, doc):
        """Return a new :class:`aeidon.Parser` instance."""
        re_tag = self.get_markup_tag_regex(doc)
        clean_func = self.get_markup_clean_func(doc)
        return aeidon.Parser(re_tag, clean_func)

    def get_text_length_require(self, index, doc):
        assert 0 <= index < len(self.subtitles)

    @aeidon.deco.export
    def get_text_length(self, index, doc):
        """Return the amount of characters in text excluding markup."""
        text = self.subtitles[index].get_text(doc)
        re_tag = self.get_markup_tag_regex(doc)
        if re_tag is not None:
            text = re_tag.sub("", text)
        return len(text)

    @aeidon.deco.export
    def get_text_signal(self, doc):
        """Return the ``texts-changed`` signal corresponding to `doc`."""
        if doc == aeidon.documents.MAIN:
            return "main-texts-changed"
        if doc == aeidon.documents.TRAN:
            return "translation-texts-changed"
        raise ValueError("Invalid document: %s" % repr(doc))

    @aeidon.deco.export
    def new_revertable_action(self, register):
        """Return a new :class:`aeidon.RevertableAction` instance."""
        action = aeidon.RevertableAction()
        action.register = register
        return action

    @aeidon.deco.export
    def new_subtitle(self):
        """Return a new :class:`aeidon.Subtitle` instance."""
        return aeidon.Subtitle(self.get_mode(), self.framerate)

    def new_temp_file_require(self, doc, encoding=None):
        assert self.get_file(doc) is not None
        if encoding is not None:
            assert aeidon.encodings.is_valid_code(encoding)

    def new_temp_file_ensure(self, value, doc, encoding=None):
        assert os.path.isfile(value)

    @aeidon.deco.export
    def new_temp_file(self, doc, encoding=None):
        """Return path to a new temporary file with subtitles from `doc`.

        Raise :exc:`IOError` if writing to temporary file fails.
        Raise :exc:`UnicodeError` if encoding temporary file fails.
        """
        sfile = self.get_file(doc)
        if sfile is None and doc == aeidon.documents.TRAN:
            # For an unsaved translation document,
            # fall back to main document's properties.
            sfile = self.get_file(aeidon.documents.MAIN)
        if sfile is not None:
            path = aeidon.temp.create(sfile.format.extension)
            encoding = encoding or sfile.encoding
            sfile = aeidon.files.new(sfile.format, path, encoding)
            sfile.copy_from(self.get_file(doc))
        else:
            # If no saved document to pull properties from,
            # fall back to SubRip format and UTF-8 encoding.
            path = aeidon.temp.create(aeidon.formats.SUBRIP.extension)
            encoding = encoding or "utf_8"
            sfile = aeidon.files.new(aeidon.formats.SUBRIP, path, encoding)
        self.save(doc, sfile, False)
        return sfile.path



# Copyright (C) 2006-2009 Osmo Salomaa
#
# This file is part of Gaupol.
#
# Gaupol is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Gaupol is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Gaupol. If not, see <http://www.gnu.org/licenses/>.

"""Miscellaneous decorators for functions and methods."""

import aeidon
import copy
import cPickle
import functools
import time


def _dump_subtitles(subtitles):
    """Return a tuple of essential attributes of subtitles."""
    return tuple((subtitle._start,
                  subtitle._end,
                  subtitle._main_text,
                  subtitle._tran_text,
                  subtitle._framerate) for subtitle in subtitles)

def _hasattr_def(obj, name):
    """Return ``True`` if `obj` has attribute `name` defined."""
    if hasattr(obj, "__dict__"):
        return name in obj.__dict__
    return hasattr(obj, name)

def _is_method(function, args):
    """Return ``True`` if `function` to be decorated is a method.

    Decorator is required to have set an `original` attribute on the wrapped
    method pointing to the original unwrapped function.
    """
    try:
        method = getattr(args[0], function.__name__)
        return (method.original is function)
    except (IndexError, AttributeError):
        return False

def benchmark(function):
    """Decorator for benchmarking functions and methods."""
    @functools.wraps(function)
    def wrapper(*args, **kwargs):
        a = time.time()
        value = function(*args, **kwargs)
        z = time.time()
        print "%.3f %s" % (z - a, function.__name__)
        return value
    return wrapper

def contractual(function):
    """Decorator for module level functions with pre- and/or postconditions.

    `function` call will be wrapped around ``FUNCTION_NAME_require`` and
    ``FUNCTION_NAME_ensure`` calls if such functions exist. The require
    function receives the same arguments as function, the ensure function will
    in addition receive function's return value as its first argument. This is
    a debug decorator that is in use only if :data:`aeidon.debug` is ``True``.
    """
    if not aeidon.debug:
        return function
    @functools.wraps(function)
    def wrapper(*args, **kwargs):
        name = "%s_require" % function.__name__
        if name in function.func_globals:
            function.func_globals[name](*args, **kwargs)
        value = function(*args, **kwargs)
        name = "%s_ensure" % function.__name__
        if name in function.func_globals:
            function.func_globals[name](value, *args, **kwargs)
        return value
    return wrapper

def export(function):
    """Decorator for delegate functions that are exported to master."""
    function.export = True
    @functools.wraps(function)
    def wrapper(*args, **kwargs):
        return function(*args, **kwargs)
    return wrapper

def memoize(function):
    """Decorator for functions that cache their return values."""
    cache = {}
    @functools.wraps(function)
    def wrapper(*args, **kwargs):
        params = (args, kwargs)
        if _is_method(function, args):
            params = (id(args[0]), args[1:], kwargs)
        key = cPickle.dumps(params)
        if not key in cache:
            cache[key] = function(*args, **kwargs)
        return cache[key]
    wrapper.original = function
    return wrapper

def monkey_patch(obj, name):
    """Decorator for functions that change `obj`'s `name` attribute.

    Any changes done will be reverted after the function is run, i.e. `name`
    attribute is either restored to its original value or deleted, if it didn't
    originally exist. The attribute in question must be able to correctly
    handle a :func:`copy.deepcopy` operation.

    Typical use would be unit testing code under legitimately unachievable
    conditions, e.g. pseudo-testing behaviour on Windows, while not actually
    using Windows::

        @aeidon.deco.monkey_patch(sys, "platform")
        def test_do_something():
            sys.platform = "win32"
            do_something()
    """
    def outer_wrapper(function):
        @functools.wraps(function)
        def inner_wrapper(*args, **kwargs):
            has_attr_def = _hasattr_def(obj, name)
            if has_attr_def:
                attr = getattr(obj, name)
                setattr(obj, name, copy.deepcopy(attr))
            try:
                return function(*args, **kwargs)
            finally:
                if has_attr_def:
                    setattr(obj, name, attr)
                    assert getattr(obj, name) == attr
                    assert getattr(obj, name) is attr
                else:
                    delattr(obj, name)
                    assert not _hasattr_def(obj, name)
        return inner_wrapper
    return outer_wrapper

def notify_frozen(function):
    """Decorator for methods to be run in notify frozen state."""
    @functools.wraps(function)
    def wrapper(*args, **kwargs):
        frozen = args[0].freeze_notify()
        try: value = function(*args, **kwargs)
        finally: args[0].thaw_notify(frozen)
        return value
    return wrapper

def once(function):
    """Decorator for functions that cache their only return value."""
    cache = []
    @functools.wraps(function)
    def wrapper(*args, **kwargs):
        if not cache:
            cache.append(function(*args, **kwargs))
        return cache[0]
    return wrapper

def reversion_test(function):
    """Decorator for unit testing reversions of one action."""
    @functools.wraps(function)
    def wrapper(*args, **kwargs):
        project = args[0].project
        original = _dump_subtitles(project.subtitles)
        value = function(*args, **kwargs)
        changed = _dump_subtitles(project.subtitles)
        assert changed != original
        for i in range(2):
            project.undo()
            current = _dump_subtitles(project.subtitles)
            assert current == original
            project.redo()
            current = _dump_subtitles(project.subtitles)
            assert current == changed
        return value
    return wrapper

def revertable(function):
    """Decorator for revertable methods of :class:`aeidon.Project`."""
    @functools.wraps(function)
    def wrapper(*args, **kwargs):
        project = args[0]
        main_changed = project.main_changed
        tran_changed = project.tran_changed
        kwargs.setdefault("register", aeidon.registers.DO)
        register = kwargs["register"]
        if register is None:
            # Execute plain function for special-case actions
            # that are not to be pushed to the undo stack.
            return function(*args, **kwargs)
        blocked = project.block(register.signal)
        if not blocked:
            # Execute plain function for nested function calls
            # that are part of another revertable action.
            return function(*args, **kwargs)
        try: value = function(*args, **kwargs)
        finally: project.unblock(register.signal)
        project.cut_reversion_stacks()
        if ((project.main_changed != main_changed) or
            (project.tran_changed != tran_changed)):
            project.emit_action_signal(register)
        return value
    return wrapper

def silent(*exceptions):
    """Decorator for ignoring `exceptions` raised  by function.

    If no exceptions specified, ignore :exc:`Exception`.
    Return ``None`` if an exception encountered.
    """
    if not exceptions:
        exceptions = (Exception,)
    def outer_wrapper(function):
        @functools.wraps(function)
        def inner_wrapper(*args, **kwargs):
            try: return function(*args, **kwargs)
            except exceptions: return None
        return inner_wrapper
    return outer_wrapper



# Copyright (C) 2006-2009 Osmo Salomaa
#
# This file is part of Gaupol.
#
# Gaupol is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Gaupol is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Gaupol. If not, see <http://www.gnu.org/licenses/>.

import copy
import aeidon


class PuppetMaster(object):

    def __init__(self):
        self.count = 0

    def notify(self, name):
        self.count += 1


class _TestObservable(aeidon.TestCase):

    def edit_obs(self):
        raise NotImplementedError

    def setup_method(self, method):
        self.master = PuppetMaster()
        self.obs = None

    def teardown_method(self, method):
        if method.__name__ in self.__class__.__dict__:
            assert self.master.count == 1

    def test___copy__(self):
        obs_copy = copy.copy(self.obs)
        assert obs_copy == self.obs
        assert obs_copy.master is self.obs.master
        self.edit_obs()
        assert obs_copy != self.obs
        assert obs_copy.master is self.obs.master

    def test___deepcopy__(self):
        obs_copy = copy.deepcopy(self.obs)
        assert obs_copy == self.obs
        assert obs_copy.master is self.obs.master
        self.edit_obs()
        assert obs_copy != self.obs
        assert obs_copy.master is self.obs.master


class TestObservableDict(_TestObservable):

    def edit_obs(self):
        del self.obs[1]

    def setup_method(self, method):
        _TestObservable.setup_method(self, method)
        self.obs = aeidon.ObservableDict(((1, 1), (2, 2)), self.master, "")

    def test___delitem__(self):
        del self.obs[1]

    def test___setitem__(self):
        self.obs[1] = 2

    def test_clear(self):
        self.obs.clear()

    def test_pop(self):
        self.obs.pop(1)

    def test_popitem(self):
        self.obs.popitem()

    def test_setdefault(self):
        self.obs.setdefault(1, 2)

    def test_update(self):
        self.obs.update({1: 2})


class TestObservableList(_TestObservable):

    def edit_obs(self):
        self.obs.pop()

    def setup_method(self, method):
        _TestObservable.setup_method(self, method)
        self.obs = aeidon.ObservableList((1, 2, 3), self.master, "")

    def test___delitem__(self):
        del self.obs[0]

    def test___delslice__(self):
        del self.obs[0:2]

    def test___iadd__(self):
        self.obs += [4, 5]

    def test___imul__(self):
        self.obs *= 2

    def test___setitem__(self):
        self.obs[0] = 2

    def test___setslice__(self):
        self.obs[0:2] = [2, 3]

    def test_append(self):
        self.obs.append(4)

    def test_extend(self):
        self.obs.extend([4, 5])

    def test_insert(self):
        self.obs.insert(0, 0)

    def test_pop(self):
        self.obs.pop()

    def test_remove(self):
        self.obs.remove(1)

    def test_reverse(self):
        self.obs.reverse()

    def test_sort(self):
        self.obs.sort()


class TestObservableSet(_TestObservable):

    def edit_obs(self):
        self.obs.pop()

    def setup_method(self, method):
        _TestObservable.setup_method(self, method)
        self.obs = aeidon.ObservableSet((1, 2, 3), self.master, "")

    def test___iand__(self):
        self.obs &= set((1, 2))

    def test___ior__(self):
        self.obs |= set((1, 2))

    def test___isub__(self):
        self.obs -= set((1, 2))

    def test___ixor__(self):
        self.obs ^= set((1, 2))

    def test_add(self):
        self.obs.add(4)

    def test_clear(self):
        self.obs.clear()

    def test_difference_update(self):
        self.obs.difference_update(set((1, 2)))

    def test_discard(self):
        self.obs.discard(1)

    def test_intersection_update(self):
        self.obs.intersection_update(set((1, 2)))

    def test_pop(self):
        self.obs.pop()

    def test_remove(self):
        self.obs.remove(1)

    def test_symmetric_difference_update(self):
        self.obs.symmetric_difference_update(set((1, 2)))

    def test_update(self):
        self.obs.update(set((4, 5)))



# Copyright (C) 2005-2007,2010 Osmo Salomaa
#
# This file is part of Gaupol.
#
# Gaupol is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Gaupol is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Gaupol. If not, see <http://www.gnu.org/licenses/>.

"""Formatting text."""

import aeidon


class FormatAgent(aeidon.Delegate):

    """Formatting text."""

    @aeidon.deco.export
    def _on_toggle_dialogue_dashes_activate(self, *args):
        """Add or remove dialogue dashes on the selected texts."""
        page = self.get_current_page()
        rows = page.view.get_selected_rows()
        col = page.view.get_focus()[1]
        doc = page.text_column_to_document(col)
        page.project.toggle_dialogue_dashes(rows, doc)

    @aeidon.deco.export
    def _on_toggle_italicization_activate(self, *args):
        """Italicize or unitalicize the selected texts."""
        page = self.get_current_page()
        rows = page.view.get_selected_rows()
        col = page.view.get_focus()[1]
        doc = page.text_column_to_document(col)
        page.project.toggle_italicization(rows, doc)

    @aeidon.deco.export
    def _on_use_lower_case_activate(self, *args):
        """Change the selected texts to lower case."""
        page = self.get_current_page()
        rows = page.view.get_selected_rows()
        col = page.view.get_focus()[1]
        doc = page.text_column_to_document(col)
        page.project.change_case(rows, doc, "lower")

    @aeidon.deco.export
    def _on_use_sentence_case_activate(self, *args):
        """Change the selected texts to sentence case."""
        page = self.get_current_page()
        rows = page.view.get_selected_rows()
        col = page.view.get_focus()[1]
        doc = page.text_column_to_document(col)
        page.project.change_case(rows, doc, "capitalize")

    @aeidon.deco.export
    def _on_use_title_case_activate(self, *args):
        """Change the selected texts to title case."""
        page = self.get_current_page()
        rows = page.view.get_selected_rows()
        col = page.view.get_focus()[1]
        doc = page.text_column_to_document(col)
        page.project.change_case(rows, doc, "title")

    @aeidon.deco.export
    def _on_use_upper_case_activate(self, *args):
        """Change the selected texts to upper case."""
        page = self.get_current_page()
        rows = page.view.get_selected_rows()
        col = page.view.get_focus()[1]
        doc = page.text_column_to_document(col)
        page.project.change_case(rows, doc, "upper")



# Copyright (C) 2005-2007,2009 Osmo Salomaa
#
# This file is part of Gaupol.
#
# Gaupol is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Gaupol is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Gaupol. If not, see <http://www.gnu.org/licenses/>.

import aeidon


class TestError(aeidon.TestCase):

    def test_raise__error(self):
        try:
            raise aeidon.Error
        except aeidon.Error:
            pass


class TestAffirmationError(aeidon.TestCase):

    def test_raise__affirmation_error(self):
        try:
            raise aeidon.AffirmationError
        except aeidon.AffirmationError:
            pass

    def test_raise__error(self):
        try:
            raise aeidon.AffirmationError
        except aeidon.Error:
            pass


class TestFormatError(aeidon.TestCase):

    def test_raise__format_error(self):
        try:
            raise aeidon.FormatError
        except aeidon.FormatError:
            pass

    def test_raise__error(self):
        try:
            raise aeidon.FormatError
        except aeidon.Error:
            pass


class TestParseError(aeidon.TestCase):

    def test_raise__parse_error(self):
        try:
            raise aeidon.ParseError
        except aeidon.ParseError:
            pass

    def test_raise__error(self):
        try:
            raise aeidon.ParseError
        except aeidon.Error:
            pass


class TestProcessError(aeidon.TestCase):

    def test_raise__process_error(self):
        try:
            raise aeidon.ProcessError
        except aeidon.ProcessError:
            pass

    def test_raise__error(self):
        try:
            raise aeidon.ProcessError
        except aeidon.Error:
            pass



# Copyright (C) 2005-2009 Osmo Salomaa
#
# This file is part of Gaupol.
#
# Gaupol is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Gaupol is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Gaupol. If not, see <http://www.gnu.org/licenses/>.

"""SubViewer 2.0 file."""

import aeidon
import re

__all__ = ("SubViewer2",)


class SubViewer2(aeidon.SubtitleFile):

    """SubViewer 2.0 file."""

    _re_time_line = re.compile((r"^(-?\d\d:\d\d:\d\d.\d\d)"
                                r",(-?\d\d:\d\d:\d\d.\d\d)\s*$"))

    format = aeidon.formats.SUBVIEWER2
    mode = aeidon.modes.TIME

    def read(self):
        """Read file and return subtitles.

        Raise :exc:`IOError` if reading fails.
        Raise :exc:`UnicodeError` if decoding fails.
        """
        self.header = ""
        subtitles = []
        lines = self._read_lines()
        while lines[0].startswith("["):
            self.header += "\n"
            self.header += lines.pop(0)
        self.header = self.header.lstrip()
        for i, line in enumerate(lines + [""]):
            match = self._re_time_line.match(line)
            if match is None: continue
            subtitle = self._get_subtitle()
            subtitle.start = match.group(1) + "0"
            subtitle.end = match.group(2) + "0"
            text = lines[i + 1].replace("[br]", "\n")
            subtitle.main_text = text
            subtitles.append(subtitle)
        return subtitles

    def write_to_file(self, subtitles, doc, fobj):
        """Write `subtitles` from `doc` to `fobj`.

        Raise :exc:`IOError` if writing fails.
        Raise :exc:`UnicodeError` if encoding fails.
        """
        n = self.newline.value
        fobj.write(self.header.replace("\n", n))
        fobj.write(n)
        for subtitle in subtitles:
            fobj.write(n)
            start = subtitle.calc.round_time(subtitle.start_time, 2)[:-1]
            end = subtitle.calc.round_time(subtitle.end_time, 2)[:-1]
            fobj.write("%s,%s%s" % (start, end, n))
            fobj.write(subtitle.get_text(doc).replace("\n", "[br]"))
            fobj.write(n)



# Copyright (C) 2005-2008,2010 Osmo Salomaa
#
# This file is part of Gaupol.
#
# Gaupol is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Gaupol is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Gaupol. If not, see <http://www.gnu.org/licenses/>.

"""Dialog for selecting subtitle files to open."""

import aeidon
import gaupol
import gtk
import os

__all__ = ("OpenDialog",)


class OpenDialog(gaupol.FileDialog):

    """Dialog for selecting subtitle files to open."""

    _widgets = ("align_combo", "align_label", "encoding_combo")

    def __init__(self, parent, title, doc):
        """Initialize an :class:`OpenDialog` object."""
        gaupol.FileDialog.__init__(self, "open-dialog.ui")
        self._use_autodetection = aeidon.util.chardet_available()
        self._init_filters()
        self._init_encoding_combo()
        self._init_align_combo()
        self._init_values(doc)
        self.set_title(title)
        self.set_transient_for(parent)

    def _init_align_combo(self):
        """Initialize the align method combo box."""
        store = gtk.ListStore(str)
        self._align_combo.set_model(store)
        for align_method in aeidon.align_methods:
            store.append((align_method.label,))
        view = self._align_combo.get_child()
        view.set_displayed_row(0)
        renderer = gtk.CellRendererText()
        self._align_combo.pack_start(renderer, True)
        self._align_combo.add_attribute(renderer, "text", 0)

    def _init_values(self, doc):
        """Initialize default values for widgets."""
        self.set_select_multiple(doc == aeidon.documents.MAIN)
        if os.path.isdir(gaupol.conf.file.directory):
            self.set_current_folder(gaupol.conf.file.directory)
        self.set_encoding(gaupol.conf.file.encoding)
        self._align_combo.set_active(gaupol.conf.file.align_method)
        self._align_combo.props.visible = (doc == aeidon.documents.TRAN)
        self._align_label.props.visible = (doc == aeidon.documents.TRAN)

    def _on_response(self, dialog, response):
        """Save default values for widgets."""
        gaupol.conf.file.encoding = self.get_encoding()
        directory = self.get_current_folder()
        if directory is not None:
            gaupol.conf.file.directory = directory
        index = self._align_combo.get_active()
        gaupol.conf.file.align_method = aeidon.align_methods[index]



# Copyright (C) 2005-2009 Osmo Salomaa
#
# This file is part of Gaupol.
#
# Gaupol is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Gaupol is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Gaupol. If not, see <http://www.gnu.org/licenses/>.

import aeidon


class TestUtilityAgent(aeidon.TestCase):

    def setup_method(self, method):
        self.project = self.new_project()
        self.delegate = self.project.get_changed.im_self

    def test_get_all_indices(self):
        indices = self.project.get_all_indices()
        assert indices == range(len(self.project.subtitles))

    def test_get_changed__main(self):
        changed = self.project.get_changed(aeidon.documents.MAIN)
        assert changed == self.project.main_changed

    def test_get_changed__translation(self):
        changed = self.project.get_changed(aeidon.documents.TRAN)
        assert changed == self.project.tran_changed

    def test_get_changed__value_error(self):
        self.raises(ValueError, self.project.get_changed, None)

    def test_get_file__main(self):
        sfile = self.project.get_file(aeidon.documents.MAIN)
        assert sfile is self.project.main_file

    def test_get_file__translation(self):
        sfile = self.project.get_file(aeidon.documents.TRAN)
        assert sfile is self.project.tran_file

    def test_get_file__value_error(self):
        self.raises(ValueError, self.project.get_file, None)

    def test_get_format__main(self):
        format = self.project.get_format(aeidon.documents.MAIN)
        assert format == self.project.main_file.format

    def test_get_format__main_is_none(self):
        self.project.main_file = None
        format = self.project.get_format(aeidon.documents.MAIN)
        assert format is None

    def test_get_format__translation(self):
        format = self.project.get_format(aeidon.documents.TRAN)
        assert format == self.project.tran_file.format

    def test_get_format__translation_is_none(self):
        self.project.tran_file = None
        format = self.project.get_format(aeidon.documents.TRAN)
        assert format == self.project.main_file.format

    def test_get_format__value_error(self):
        self.raises(ValueError, self.project.get_format, None)

    def test_get_liner(self):
        doc = aeidon.documents.MAIN
        liner = self.project.get_liner(doc)
        assert liner.re_tag == self.project.get_markup_tag_regex(doc)

    def test_get_markup__main(self):
        markup = self.project.get_markup(aeidon.documents.MAIN)
        format = self.project.main_file.format
        assert markup == aeidon.tags.new(format)

    def test_get_markup__translation(self):
        markup = self.project.get_markup(aeidon.documents.TRAN)
        format = self.project.tran_file.format
        assert markup == aeidon.tags.new(format)

    def test_get_markup_clean_func(self):
        doc = aeidon.documents.MAIN
        clean_func = self.project.get_markup_clean_func(doc)
        assert clean_func("") == ""

    def test_get_markup_tag_regex(self):
        re_tag = self.project.get_markup_tag_regex(aeidon.documents.MAIN)
        assert re_tag is not None

    def test_get_markup_tag_regex__none(self):
        self.project.main_file = None
        re_tag = self.project.get_markup_tag_regex(aeidon.documents.MAIN)
        assert re_tag is None

    def test_get_mode__default(self):
        self.project.main_file = None
        mode = self.project.get_mode()
        assert mode == aeidon.modes.TIME

    def test_get_mode__frame(self):
        self.project.open_main(self.new_microdvd_file(), "ascii")
        mode = self.project.get_mode()
        assert mode == aeidon.modes.FRAME

    def test_get_mode__time(self):
        self.project.open_main(self.new_subrip_file(), "ascii")
        mode = self.project.get_mode()
        assert mode == aeidon.modes.TIME

    def test_get_parser(self):
        doc = aeidon.documents.MAIN
        parser = self.project.get_parser(doc)
        assert parser.re_tag == self.project.get_markup_tag_regex(doc)

    def test_get_text_length(self):
        self.project.subtitles[0].main_text = "<i>test\ntest.</i>"
        length = self.project.get_text_length(0, aeidon.documents.MAIN)
        assert length == 10

    def test_get_text_signal__main(self):
        signal = self.project.get_text_signal(aeidon.documents.MAIN)
        assert signal == "main-texts-changed"

    def test_get_text_signal__translation(self):
        signal = self.project.get_text_signal(aeidon.documents.TRAN)
        assert signal == "translation-texts-changed"

    def test_get_text_signal__value_error(self):
        self.raises(ValueError, self.project.get_text_signal, None)

    def test_new_revertable_action(self):
        register = aeidon.registers.DO
        action = self.project.new_revertable_action(register)
        assert action.register == register

    def test_new_subtitle(self):
        subtitle = self.project.new_subtitle()
        assert subtitle.mode == self.project.main_file.mode
        assert subtitle.framerate == self.project.framerate

    def test_new_temp_file__main(self):
        self.project.new_temp_file(aeidon.documents.MAIN)
        self.project.new_temp_file(aeidon.documents.MAIN, "ascii")

    def test_new_temp_file__translation(self):
        self.project.new_temp_file(aeidon.documents.TRAN)
        self.project.new_temp_file(aeidon.documents.TRAN, "ascii")



# Copyright (C) 2007-2009 Osmo Salomaa
#
# This file is part of Gaupol.
#
# Gaupol is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Gaupol is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Gaupol. If not, see <http://www.gnu.org/licenses/>.

"""Data store and basic position manipulation of a single subtitle."""

import aeidon
import copy

__all__ = ("Subtitle",)


class Subtitle(object):

    """Data store and basic position manipulation of a single subtitle.

    :ivar start: Start position in native units (read/write)
    :ivar start_time: Start time as string (read)
    :ivar start_frame: Start frame as integer (read)
    :ivar start_seconds: Start seconds as float (read)
    :ivar end: End position in native units (read/write)
    :ivar end_time: End time as string (read)
    :ivar end_frame: End frame as integer (read)
    :ivar end_seconds: End seconds as float (read)
    :ivar duration: Duration in native units (read/write)
    :ivar duration_time: Duration in time as string (read)
    :ivar duration_frame: Duration in frames as integer (read)
    :ivar duration_seconds: Duration in seconds as float (read)
    :ivar main_text: Main text (read/write)
    :ivar tran_text: Translation text (read/write)
    :ivar calc: :class:`aeidon.Calculator` instance used (read/write)
    :ivar framerate: :attr:`aeidon.framerates` item (read/write)
    :ivar mode: :attr:`aeidon.modes` item (read/write)

    Positions can be set as strings for times, integers for frames or floats
    for seconds. Positions are saved internally in only one mode.

    Additional format-specific attributes are kept under separate containers,
    e.g. ``ssa`` for Sub Station Alpha formats, accessed as ``subtitle.ssa.*``.
    These containers are lazily created upon first use in order to avoid slow
    instantiation and excessive memory use when handling simpler formats.
    """

    def __cmp__(self, other):
        """Compare start positions in this subtitle's mode.

        Return 1 if this subtitle appears later, 0 if they appear at the same
        time and -1 if `other` appears later.
        """
        if self._mode == aeidon.modes.TIME:
            times = (self._start, other.start_time)
            return self.calc.compare_times(*times)
        if self._mode == aeidon.modes.FRAME:
            return cmp(self._start, other.start_frame)
        raise ValueError("Invalid mode: %s" % repr(self._mode))

    def __getattr__(self, name):
        """Return lazily instantiated format-specific attribute container."""
        if name in (x.container for x in aeidon.formats):
            # Lazily instantiate a new container.
            container = aeidon.containers.new(name)
            object.__setattr__(self, name, container)
            return container
        raise AttributeError("Invalid container name: %s" % repr(name))

    def __init__(self, mode=None, framerate=None):
        """Initialize a :class:`Subtitle` object."""
        self._start = "00:00:00.000"
        self._end = "00:00:00.000"
        self._main_text = ""
        self._tran_text = ""
        self._mode = mode or aeidon.modes.TIME
        self._framerate = framerate or aeidon.framerates.FPS_23_976
        self.calc = aeidon.Calculator(self._framerate)

    def _convert_position(self, value):
        """Return `value` of position in correct mode."""
        if isinstance(value, basestring):
            if self._mode == aeidon.modes.TIME:
                return value
            if self._mode == aeidon.modes.FRAME:
                return self.calc.time_to_frame(value)
        if isinstance(value, int):
            if self._mode == aeidon.modes.TIME:
                return self.calc.frame_to_time(value)
            if self._mode == aeidon.modes.FRAME:
                return value
        if isinstance(value, float):
            if self._mode == aeidon.modes.TIME:
                return self.calc.seconds_to_time(value)
            if self._mode == aeidon.modes.FRAME:
                return self.calc.seconds_to_frame(value)
        raise ValueError("Invalid type for value: %s" % repr(type(value)))

    def convert_framerate(self, framerate):
        """Set framerate and convert positions to it."""
        coefficient = framerate.value / self._framerate.value
        self.framerate = framerate
        if self._mode == aeidon.modes.TIME:
            self.start = self.start_seconds / coefficient
            self.end = self.end_seconds / coefficient
        if self._mode == aeidon.modes.FRAME:
            self.start = int(round(coefficient * self.start_frame, 0))
            self.end = int(round(coefficient * self.end_frame, 0))

    def copy(self):
        """Return a new subtitle instance with the same values."""
        subtitle = Subtitle(self._mode, self._framerate)
        subtitle._start = self._start
        subtitle._end = self._end
        subtitle._main_text = self._main_text
        subtitle._tran_text = self._tran_text
        # Copy all containers that have been instantiated.
        containers = (x.container for x in aeidon.formats)
        for name in (set(dir(self)) & set(containers)):
            container = copy.deepcopy(getattr(self, name))
            setattr(subtitle, name, container)
        return subtitle

    @property
    def duration(self):
        """Return duration in correct mode."""
        if self._mode == aeidon.modes.TIME:
            return self.duration_time
        if self._mode == aeidon.modes.FRAME:
            return self.duration_frame
        raise ValueError("Invalid mode: %s" % repr(self._mode))

    @duration.setter
    def duration(self, value):
        """Set duration from `value`."""
        value = self._convert_position(value)
        if self._mode == aeidon.modes.TIME:
            times = (self.start_time, value)
            self._end = self.calc.add_times(*times)
        if self._mode == aeidon.modes.FRAME:
            self._end = self.start_frame + value

    @property
    def duration_frame(self):
        """Return duration as frames."""
        bounds = (self.start_frame, self.end_frame)
        return self.calc.get_frame_duration(*bounds)

    @property
    def duration_seconds(self):
        """Return duration as seconds."""
        time = self.duration_time
        return self.calc.time_to_seconds(time)

    @property
    def duration_time(self):
        """Return duration as time."""
        bounds = (self.start_time, self.end_time)
        return self.calc.get_time_duration(*bounds)

    @property
    def end(self):
        """Return end position in correct mode."""
        return self._end

    @end.setter
    def end(self, value):
        """Set end position from `value`."""
        self._end = self._convert_position(value)

    @property
    def end_frame(self):
        """Return end position as frames."""
        if self._mode == aeidon.modes.TIME:
            return self.calc.time_to_frame(self._end)
        if self._mode == aeidon.modes.FRAME:
            return self._end
        raise ValueError("Invalid mode: %s" % repr(self._mode))

    @property
    def end_seconds(self):
        """Return end position as seconds."""
        time = self.end_time
        return self.calc.time_to_seconds(time)

    @property
    def end_time(self):
        """Return end position as time."""
        if self._mode == aeidon.modes.TIME:
            return self._end
        if self._mode == aeidon.modes.FRAME:
            return self.calc.frame_to_time(self._end)
        raise ValueError("Invalid mode: %s" % repr(self._mode))

    @property
    def framerate(self):
        """Return framerate."""
        return self._framerate

    @framerate.setter
    def framerate(self, value):
        """Set framerate from `value`."""
        self._framerate = value
        self.calc = aeidon.Calculator(value)

    def get_duration(self, mode):
        """Return duration in `mode`."""
        if mode == aeidon.modes.TIME:
            return self.duration_time
        if mode == aeidon.modes.FRAME:
            return self.duration_frame
        raise ValueError("Invalid mode: %s" % repr(mode))

    def get_end(self, mode):
        """Return end position in `mode`."""
        if mode == aeidon.modes.TIME:
            return self.end_time
        if mode == aeidon.modes.FRAME:
            return self.end_frame
        raise ValueError("Invalid mode: %s" % repr(mode))

    def get_start(self, mode):
        """Return start position in `mode`."""
        if mode == aeidon.modes.TIME:
            return self.start_time
        if mode == aeidon.modes.FRAME:
            return self.start_frame
        raise ValueError("Invalid mode: %s" % repr(mode))

    def get_text(self, doc):
        """Return text corresponding to `doc`."""
        if doc == aeidon.documents.MAIN:
            return self._main_text
        if doc == aeidon.documents.TRAN:
            return self._tran_text
        raise ValueError("Invalid document: %s" % repr(doc))

    def has_container(self, name):
        """Return ``True`` if container has been instantiated."""
        return name in dir(self)

    @property
    def main_text(self):
        """Return main text."""
        return self._main_text

    @main_text.setter
    def main_text(self, value):
        """Set main text from `value`."""
        self._main_text = unicode(value)

    @property
    def mode(self):
        """Return current position mode."""
        return self._mode

    @mode.setter
    def mode(self, mode):
        """Set current position mode."""
        if mode == aeidon.modes.TIME:
            self._start = self.start_time
            self._end = self.end_time
        if mode == aeidon.modes.FRAME:
            self._start = self.start_frame
            self._end = self.end_frame
        self._mode = mode

    def scale_positions(self, value):
        """Multiply start and end positions by `value`."""
        if self._mode == aeidon.modes.TIME:
            self.start = self.start_seconds * value
            self.end = self.end_seconds * value
        if self._mode == aeidon.modes.FRAME:
            self.start = int(round(self._start * value, 0))
            self.end = int(round(self._end * value, 0))

    def set_text(self, doc, value):
        """Set text corresponding to `doc` to `value`."""
        if doc == aeidon.documents.MAIN:
            self.main_text = value
        if doc == aeidon.documents.TRAN:
            self.tran_text = value

    def shift_positions(self, value):
        """Add `value` to start and end positions."""
        if isinstance(value, basestring):
            start = self.calc.add_times(self.start_time, value)
            end = self.calc.add_times(self.end_time, value)
            self.start = start
            self.end = end
        elif isinstance(value, int):
            start = self.start_frame + value
            end = self.end_frame + value
            self.start = start
            self.end = end
        elif isinstance(value, float):
            start = self.start_seconds + value
            end = self.end_seconds + value
            self.start = start
            self.end = end

    @property
    def start(self):
        """Return start position in correct mode."""
        return self._start

    @start.setter
    def start(self, value):
        """Set start position from `value`."""
        self._start = self._convert_position(value)

    @property
    def start_frame(self):
        """Return start position as frames."""
        if self._mode == aeidon.modes.TIME:
            return self.calc.time_to_frame(self._start)
        if self._mode == aeidon.modes.FRAME:
            return self._start
        raise ValueError("Invalid mode: %s" % repr(self._mode))

    @property
    def start_seconds(self):
        """Return start position as seconds."""
        time = self.start_time
        return self.calc.time_to_seconds(time)

    @property
    def start_time(self):
        """Return start position as time."""
        if self._mode == aeidon.modes.TIME:
            return self._start
        if self._mode == aeidon.modes.FRAME:
            return self.calc.frame_to_time(self._start)
        raise ValueError("Invalid mode: %s" % repr(self._mode))

    @property
    def tran_text(self):
        """Return translation text."""
        return self._tran_text

    @tran_text.setter
    def tran_text(self, value):
        """Set translation text from `value`."""
        self._tran_text = unicode(value)



# Copyright (C) 2005-2008,2010 Osmo Salomaa
#
# This file is part of Gaupol.
#
# Gaupol is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Gaupol is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Gaupol. If not, see <http://www.gnu.org/licenses/>.

"""Text formatting actions for :class:`gaupol.Application`."""

import aeidon
import gaupol
import gtk
_ = aeidon.i18n._


class ShowCaseMenuAction(gaupol.MenuAction):

    """Show the case format menu."""

    def __init__(self):
        """Initialize a :class:`ShowCaseMenuAction` object."""
        gaupol.MenuAction.__init__(self, "show_case_menu")
        self.props.label = _("Ca_se")
        self.action_group = "main-unsafe"

    def _affirm_doable(self, application, page):
        """Raise :exc:`aeidon.AffirmationError` if action cannot be done."""
        aeidon.util.affirm(page is not None)
        aeidon.util.affirm(page.view.get_selected_rows())
        col = page.view.get_focus()[1]
        aeidon.util.affirm(col is not None)
        aeidon.util.affirm(page.view.is_text_column(col))


class ToggleDialogDashesAction(gaupol.Action):

    """Add or remove dialogue dashes on the selected texts."""

    def __init__(self):
        """Initialize a :class:`ToggleDialogDashesAction` object."""
        gaupol.Action.__init__(self, "toggle_dialogue_dashes")
        self.props.label = _("_Dialogue")
        self.props.tooltip = _("Add or remove dialogue dashes "
                               "on the selected texts")

        self.accelerator = "D"
        self.action_group = "main-unsafe"

    def _affirm_doable(self, application, page):
        """Raise :exc:`aeidon.AffirmationError` if action cannot be done."""
        aeidon.util.affirm(page is not None)
        aeidon.util.affirm(page.view.get_selected_rows())
        col = page.view.get_focus()[1]
        aeidon.util.affirm(col is not None)
        aeidon.util.affirm(page.view.is_text_column(col))


class ToggleItalicizationAction(gaupol.Action):

    """Italicize or unitalicize the selected texts."""

    def __init__(self):
        """Initialize a :class:`ToggleItalicizationAction` object."""
        gaupol.Action.__init__(self, "toggle_italicization")
        self.props.label = _("_Italic")
        self.props.stock_id = gtk.STOCK_ITALIC
        self.props.tooltip = _("Italicize or unitalicize the selected texts")
        self.accelerator = "I"
        self.action_group = "main-unsafe"

    def _affirm_doable(self, application, page):
        """Raise :exc:`aeidon.AffirmationError` if action cannot be done."""
        aeidon.util.affirm(page is not None)
        aeidon.util.affirm(page.view.get_selected_rows())
        col = page.view.get_focus()[1]
        aeidon.util.affirm(col is not None)
        aeidon.util.affirm(page.view.is_text_column(col))
        doc = page.text_column_to_document(col)
        markup = page.project.get_markup(doc)
        aeidon.util.affirm(markup is not None)
        aeidon.util.affirm(markup.italic_tag is not None)


class UseLowerCaseAction(gaupol.Action):

    """Change the selected texts to lower case."""

    def __init__(self):
        """Initialize a :class:`UseLowerCaseAction` object."""
        gaupol.Action.__init__(self, "use_lower_case")
        self.props.label = _("_Lower")
        self.props.tooltip = _("Change the selected texts to lower case")
        self.action_group = "main-unsafe"

    def _affirm_doable(self, application, page):
        """Raise :exc:`aeidon.AffirmationError` if action cannot be done."""
        aeidon.util.affirm(page is not None)
        aeidon.util.affirm(page.view.get_selected_rows())
        col = page.view.get_focus()[1]
        aeidon.util.affirm(col is not None)
        aeidon.util.affirm(page.view.is_text_column(col))


class UseSentenceCaseAction(gaupol.Action):

    """Change the selected texts to sentence case."""

    def __init__(self):
        """Initialize a :class:`UseSentenceCaseAction` object."""
        gaupol.Action.__init__(self, "use_sentence_case")
        self.props.label = _("_Sentence")
        self.props.tooltip = _("Change the selected texts to sentence case")
        self.action_group = "main-unsafe"

    def _affirm_doable(self, application, page):
        """Raise :exc:`aeidon.AffirmationError` if action cannot be done."""
        aeidon.util.affirm(page is not None)
        aeidon.util.affirm(page.view.get_selected_rows())
        col = page.view.get_focus()[1]
        aeidon.util.affirm(col is not None)
        aeidon.util.affirm(page.view.is_text_column(col))


class UseTitleCaseAction(gaupol.Action):

    """Change the selected texts to title case."""

    def __init__(self):
        """Initialize a :class:`UseTitleCaseAction` object."""
        gaupol.Action.__init__(self, "use_title_case")
        self.props.label = _("_Title")
        self.props.tooltip = _("Change the selected texts to title case")
        self.action_group = "main-unsafe"

    def _affirm_doable(self, application, page):
        """Raise :exc:`aeidon.AffirmationError` if action cannot be done."""
        aeidon.util.affirm(page is not None)
        aeidon.util.affirm(page.view.get_selected_rows())
        col = page.view.get_focus()[1]
        aeidon.util.affirm(col is not None)
        aeidon.util.affirm(page.view.is_text_column(col))


class UseUpperCaseAction(gaupol.Action):

    """Change the selected texts to upper case."""

    def __init__(self):
        """Initialize a :class:`UseUpperCaseAction` object."""
        gaupol.Action.__init__(self, "use_upper_case")
        self.props.label = _("_Upper")
        self.props.tooltip = _("Change the selected texts to upper case")
        self.action_group = "main-unsafe"

    def _affirm_doable(self, application, page):
        """Raise :exc:`aeidon.AffirmationError` if action cannot be done."""
        aeidon.util.affirm(page is not None)
        aeidon.util.affirm(page.view.get_selected_rows())
        col = page.view.get_focus()[1]
        aeidon.util.affirm(col is not None)
        aeidon.util.affirm(page.view.is_text_column(col))


__all__ = tuple(filter(lambda x: x.endswith("Action"), dir()))



# Copyright (C) 2005-2007,2009 Osmo Salomaa
#
# This file is part of Gaupol.
#
# Gaupol is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Gaupol is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Gaupol. If not, see <http://www.gnu.org/licenses/>.

"""String and regular expression finder and replacer."""

import aeidon
import re

__all__ = ("Finder",)


class Finder(object):

    """String and regular expression finder and replacer.

    :ivar ignore_case: ``True`` to ignore case when finding matches
    :ivar match: Regular expression object for the latest match of pattern
    :ivar match_span: Tuple of start and end position for match
    :ivar pattern: String or regular expression object to find
    :ivar pos: Current offset from the beginning of the text
    :ivar replacement: Plain- or regular expression replacement string
    :ivar text: Target text to find matches of pattern in
    """

    __metaclass__ = aeidon.Contractual

    def __init__(self):
        """Initialize a Finder object."""
        self.ignore_case = False
        self.match = None
        self.match_span = None
        self.pattern = None
        self.pos = None
        self.replacement = None
        self.text = None

    def _invariant(self):
        if self.pos is not None:
            assert (0 <= self.pos <= len(self.text))

    def next_require(self):
        assert self.pattern is not None

    def next_ensure(self, value):
        for pos in value:
            assert (0 <= pos <= len(self.text))

    def next(self):
        """Find the next match of pattern.

        Raise :exc:`StopIteration` if no next match found.
        Return tuple of match start, end position.
        """
        if isinstance(self.pattern, basestring):
            text = self.text
            pattern = self.pattern
            if self.ignore_case:
                text = text.lower()
                pattern = pattern.lower()
            try:
                index = text.index(pattern, self.pos)
            except ValueError:
                raise StopIteration
            self.match_span = (index, index + len(pattern))
        else: # Regular expression
            match = self.pattern.search(self.text, self.pos)
            if match is None:
                raise StopIteration
            # Avoid getting stuck with zero-length regular expressions.
            if match.span() == self.match_span == (self.pos, self.pos):
                if self.pos == len(self.text):
                    raise StopIteration
                self.pos += 1
                return self.next()
            self.match = match
            self.match_span = match.span()

        self.pos = self.match_span[1]
        return self.match_span

    def previous_require(self):
        assert self.pattern is not None

    def previous_ensure(self, value):
        for pos in value:
            assert (0 <= pos <= len(self.text))

    def previous(self):
        """Find the previous match of pattern.

        Raise :exc:`StopIteration` if no previous match found.
        Return tuple of match start, end position.
        """
        if isinstance(self.pattern, basestring):
            text = self.text
            pattern = self.pattern
            if self.ignore_case:
                text = text.lower()
                pattern = pattern.lower()
            try:
                index = text.rindex(pattern, 0, self.pos)
            except ValueError:
                raise StopIteration
            self.match_span = (index, index + len(pattern))
        else: # Regular expression
            iterator = self.pattern.finditer(self.text)
            match = None
            while True:
                try:
                    candidate = iterator.next()
                    if candidate.end() > self.pos:
                        raise StopIteration
                except StopIteration:
                    break
                match = candidate
            if match is None:
                raise StopIteration
            # Avoid getting stuck with zero-length regular expressions.
            if match.span() == self.match_span == (self.pos, self.pos):
                if self.pos == 0:
                    raise StopIteration
                self.pos -= 1
                return self.previous()
            self.match = match
            self.match_span = match.span()

        self.pos = self.match_span[0]
        return self.match_span

    def replace_require(self, next=True):
        assert self.match_span is not None
        assert self.pattern is not None
        assert self.replacement is not None

    def replace(self, next=True):
        """Replace the current match of pattern.

        `next` should be ``True`` to finish at end of match, ``False`` for
        beginning. Raise :exc:`re.error` if bad replacement.
        """
        a, z = self.match_span
        orig_length = len(self.text)
        replacement = self.replacement
        if not isinstance(self.pattern, basestring):
            replacement = self.match.expand(self.replacement)
        self.text = self.text[:a] + replacement + self.text[z:]

        shift = len(self.text) - orig_length
        self.pos = ((z + shift) if next else a)

        # Adapt match span to new text length
        # to avoid getting stuck with zero-length regular expressions.
        if next and (self.match_span[0] == self.match_span[1]):
            self.match_span = (self.pos, self.pos)

    def replace_all_require(self):
        assert self.pattern is not None
        assert self.replacement is not None

    def replace_all(self):
        """Replace all occurences of pattern.

        Raise :exc:`re.error` if bad replacement.
        Return the amount of substitutions made.
        """
        self.pos = 0
        self.match = None
        self.match_span = None
        count = 0
        while True:
            try:
                self.next()
            except StopIteration:
                self.pos = len(self.text)
                self.match_span = None
                break
            self.replace()
            count += 1
        return count

    def set_regex(self, pattern, flags=0, default_flags=None):
        """Set and use regular expression as pattern.

        If `default_flags` is ``None``, all of ``DOTALL``, ``MULTILINE`` and
        ``UNICODE`` are used. ``IGNORECASE`` is automatically added to flags if
        :attr:`ignore_case` is ``True``.

        Raise :exc:`re.error` if bad pattern.
        """
        if self.ignore_case:
            flags = flags | re.IGNORECASE
        if default_flags is None:
            flags = flags | re.DOTALL | re.MULTILINE | re.UNICODE
        self.pattern = re.compile(pattern, flags)

    def set_text(self, text, next=True):
        """Set the target text to search in and reset position.

        `next` should be ``True`` to start at beginning, ``False`` for end.
        """
        self.text = text
        self.match = None
        self.match_span = None
        self.pos = (0 if next else len(text))



# Copyright (C) 2005-2010 Osmo Salomaa
#
# This file is part of Gaupol.
#
# Gaupol is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Gaupol is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Gaupol. If not, see <http://www.gnu.org/licenses/>.

"""Saving documents."""

import aeidon
import gaupol
import gtk
import os
_ = aeidon.i18n._


class SaveAgent(aeidon.Delegate):

    """Saving documents."""

    @aeidon.deco.export
    def _on_save_all_documents_activate(self, *args):
        """Save all open documents."""
        silent = aeidon.deco.silent(gaupol.Default)
        save_main = silent(self.save_main)
        save_tran = silent(self.save_translation)
        for page in self.pages:
            save_main(page)
            if page.project.tran_changed is not None:
                save_tran(page)
        self.flash_message(_("Saved all open documents"))
        self.update_gui()

    @aeidon.deco.export
    def _on_save_all_documents_as_activate(self, *args):
        """Save all open documents with different properties."""
        dialog = gaupol.MultiSaveDialog(self.window, self)
        gaupol.util.flash_dialog(dialog)
        self.update_gui()

    @aeidon.deco.export
    @aeidon.deco.silent(gaupol.Default)
    def _on_save_main_document_activate(self, *args):
        """Save the current main document."""
        page = self.get_current_page()
        self.save_main(page)
        self.update_gui()

    @aeidon.deco.export
    @aeidon.deco.silent(gaupol.Default)
    def _on_save_main_document_as_activate(self, *args):
        """Save the current main document with a different name."""
        page = self.get_current_page()
        self.save_main_as(page)
        self.update_gui()

    @aeidon.deco.export
    @aeidon.deco.silent(gaupol.Default)
    def _on_save_translation_document_activate(self, *args):
        """Save the current translation document."""
        page = self.get_current_page()
        self.save_translation(page)
        self.update_gui()

    @aeidon.deco.export
    @aeidon.deco.silent(gaupol.Default)
    def _on_save_translation_document_as_activate(self, *args):
        """Save the current translation document with a different name."""
        page = self.get_current_page()
        self.save_translation_as(page)
        self.update_gui()

    def _save_document(self, page, doc, sfile=None):
        """Save `doc` of `page` to `sfile`.

        Raise :exc:`gaupol.Default` if saving failed.
        """
        sfile = sfile or page.project.get_file(doc)
        gaupol.util.set_cursor_busy(self.window)
        try: return page.project.save(doc, sfile)
        except IOError as (no, message):
            gaupol.util.set_cursor_normal(self.window)
            basename = os.path.basename(sfile.path)
            self._show_io_error_dialog(basename, message)
        except UnicodeError:
            gaupol.util.set_cursor_normal(self.window)
            basename = os.path.basename(sfile.path)
            self._show_encoding_error_dialog(basename, sfile.encoding)
        finally:
            gaupol.util.set_cursor_normal(self.window)
        raise gaupol.Default

    def _select_file(self, title, sfile=None):
        """Select a file and return a :class:`aeidon.SubtitleFile`.

        Raise :exc:`gaupol.Default` if cancelled.
        """
        gaupol.util.set_cursor_busy(self.window)
        dialog = gaupol.SaveDialog(self.window, title)
        if sfile is not None:
            dialog.set_name(sfile.path)
            dialog.set_format(sfile.format)
            dialog.set_encoding(sfile.encoding)
            dialog.set_newline(sfile.newline)
        gaupol.util.set_cursor_normal(self.window)
        response = gaupol.util.run_dialog(dialog)
        format = dialog.get_format()
        path = dialog.get_filename()
        encoding = dialog.get_encoding()
        newline = dialog.get_newline()
        dialog.destroy()
        if response != gtk.RESPONSE_OK:
            raise gaupol.Default
        gaupol.util.iterate_main()
        return aeidon.files.new(format, path, encoding, newline)

    def _show_encoding_error_dialog(self, basename, codec):
        """Show an error dialog after failing to encode file."""
        codec = aeidon.encodings.code_to_name(codec)
        title = _('Failed to encode file "%(basename)s" '
            'with codec "%(codec)s"') % locals()
        message = _("Please try to save the file with "
            "a different character encoding.")
        dialog = gaupol.ErrorDialog(self.window, title, message)
        dialog.add_button(gtk.STOCK_OK, gtk.RESPONSE_OK)
        gaupol.util.flash_dialog(dialog)

    def _show_io_error_dialog(self, basename, message):
        """Show an error dialog after failing to write file."""
        title = _('Failed to save file "%s"') % basename
        dialog = gaupol.ErrorDialog(self.window, title, message)
        dialog.add_button(gtk.STOCK_OK, gtk.RESPONSE_OK)
        gaupol.util.flash_dialog(dialog)

    @aeidon.deco.export
    def save_main(self, page):
        """Save the main document of `page`.

        Raise :exc:`gaupol.Default` if cancelled or saving failed.
        """
        if (page.project.main_file is None or
            page.project.main_file.path is None or
            page.project.main_file.encoding is None):
            return self.save_main_as(page)

        self._save_document(page, aeidon.documents.MAIN)
        self.flash_message(_("Saved main document"))

    @aeidon.deco.export
    def save_main_as(self, page, sfile=None):
        """Save the main document of `page` to a selected file.

        If `sfile` is ``None`` show a filechooser dialog.
        Raise :exc:`gaupol.Default` if cancelled or saving failed.
        """
        if sfile is None:
            sfile = self._select_file(_("Save As"),
                                      page.project.main_file)

        self._save_document(page, aeidon.documents.MAIN, sfile)
        self.add_to_recent_files(page.project.main_file.path,
                                 page.project.main_file.format,
                                 aeidon.documents.MAIN)

        self.flash_message(_('Saved main document as "%s"')
                           % os.path.basename(page.project.main_file.path))

    @aeidon.deco.export
    def save_translation(self, page):
        """Save the translation document of `page`.

        Raise :exc:`gaupol.Default` if cancelled or saving failed.
        """
        if (page.project.tran_file is None or
            page.project.tran_file.path is None or
            page.project.tran_file.encoding is None):
            return self.save_translation_as(page)

        self._save_document(page, aeidon.documents.TRAN)
        self.flash_message(_("Saved translation document"))

    @aeidon.deco.export
    def save_translation_as(self, page, sfile=None):
        """Save the translation document of `page` to a selected file.

        If `sfile` is ``None`` show a filechooser dialog.
        Raise :exc:`gaupol.Default` if cancelled or saving failed.
        """
        if sfile is None:
            sfile = self._select_file(_("Save Translation As"),
                                      page.project.tran_file)

        self._save_document(page, aeidon.documents.TRAN, sfile)
        self.add_to_recent_files(page.project.tran_file.path,
                                 page.project.tran_file.format,
                                 aeidon.documents.TRAN)

        self.flash_message(_('Saved translation document as "%s"')
                           % os.path.basename(page.project.tran_file.path))



# Copyright (C) 2005-2009 Osmo Salomaa
#
# This file is part of Gaupol.
#
# Gaupol is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Gaupol is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Gaupol. If not, see <http://www.gnu.org/licenses/>.

"""Base class for GTK+ unit test cases."""

import aeidon
import gaupol

__all__ = ("TestCase",)


class TestCase(aeidon.TestCase):

    """Base class for GTK+ unit test cases.

    Unit tests are designed to be run with ``py.test``, ``nose`` or something
    compatible. Tests should use plain ``assert`` statements to allow multiple
    different tools to be used to run the tests.
    """

    def new_application(self):
        """Return a new application with two open pages."""
        application = gaupol.Application()
        application.add_page(self.new_page())
        application.window.show()
        return application

    def new_page(self):
        """Return a new page with two open documents."""
        page = gaupol.Page()
        page.project.open_main(self.new_subrip_file(), "ascii")
        page.project.open_translation(self.new_microdvd_file(), "ascii", False)
        return page

    def teardown_method(self, method):
        """Remove state set for executing tests in `method`."""
        # pylint: disable=E1101
        gaupol.util.iterate_main()
        for name in ("assistant", "dialog", "window"):
            if hasattr(self, name):
                getattr(self, name).destroy()
        if hasattr(self, "application"):
            for page in self.application.pages:
                self.application.close(page, False)
            self.application.window.hide()
            self.application.window.destroy()
        gaupol.util.iterate_main()
        gaupol.conf.restore_defaults()



# Copyright (C) 2008,2010 Osmo Salomaa
#
# This file is part of Gaupol.
#
# Gaupol is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Gaupol is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Gaupol. If not, see <http://www.gnu.org/licenses/>.

"""Finding activating, storing and deactivating extensions."""

import aeidon
import gaupol
import inspect
import os
import re
import sys
import traceback

__all__ = ("ExtensionManager", "ExtensionMetadata",)


class ExtensionManager(object):

    """Finding activating, storing and deactivating extensions.

    :ivar _active: Dictionary mapping module names to extensions
    :ivar _metadata: Dictionary mapping module names to metadata items
    :ivar _dependants: Dictionary mapping module names to a list of others
    :ivar _inferior: List of modules activated to satisfy a dependency
    """

    __metaclass__ = aeidon.Contractual
    _global_dir = os.path.join(aeidon.DATA_DIR, "extensions")
    _local_dir = os.path.join(aeidon.DATA_HOME_DIR, "extensions")
    _re_comment = re.compile(r"#.*$")

    def __init__(self, application):
        """Initialize an :class:`ExtensionManager` object."""
        self.application = application
        self._active = {}
        self._dependants = {}
        self._inferior = []
        self._metadata = {}

    def _find_extensions_in_directory(self, directory):
        """Find all extensions in `directory` and parse their metadata."""
        def is_metadata_file(path):
            return (path.endswith(".gaupol-extension") or
                    path.endswith(".gaupol-extension.in"))

        for (root, dirs, files) in os.walk(directory):
            files = filter(is_metadata_file, files)
            for name in [x for x in files if x.endswith(".in")]:
                # If both untranslated and translated metadata files are found,
                # load extension only from the translated one.
                if name[:-3] in files:
                    files.remove(name)
            for name in files:
                path = os.path.abspath(os.path.join(root, name))
                self._parse_metadata(path)

    def _parse_metadata(self, path):
        """Parse extension metadata file at `path`."""
        try: lines = aeidon.util.readlines(path, "utf_8", None)
        except UnicodeError: # Metadata file must be UTF-8.
            return aeidon.util.print_read_unicode(sys.exc_info(),
                                                  path,
                                                  "utf_8")

        lines = [self._re_comment.sub("", x) for x in lines]
        lines = [x.strip() for x in lines]
        metadata = ExtensionMetadata()
        metadata.path = path
        for line in filter(None, lines):
            if line.startswith("["): continue
            name, value = unicode(line).split("=", 1)
            name = (name[1:] if name.startswith("_") else name)
            metadata.set_field(name, value)
        self._store_metadata(path, metadata)

    def _store_metadata(self, path, metadata):
        """Store metadata to instance variables after validation."""
        def discard_extension(name):
            print ("Field '%s' missing in file '%s', dicarding extension"
                   % (name, path))

        if not metadata.has_field("GaupolVersion"):
            return discard_extension("GaupolVersion")
        if not metadata.has_field("Module"):
            return discard_extension("Module")
        if not metadata.has_field("Name"):
            return discard_extension("Name")
        if not metadata.has_field("Description"):
            return discard_extension("Description")
        module = metadata.get_field("Module")
        self._metadata[module] = metadata

    def find_extensions(self):
        """Find all extensions and parse their metadata files."""
        self._find_extensions_in_directory(self._global_dir)
        self._find_extensions_in_directory(self._local_dir)

    def get_metadata(self, module):
        """Return :class:`ExtensionMetadata` instance for module."""
        return self._metadata[module]

    def get_modules(self):
        """Return a list of all extension module names."""
        return self._metadata.keys()

    def has_help_require(self, module):
        assert module in self._active
        assert module in self._metadata

    def has_help(self, module):
        """Return ``True`` if extension can display documentation."""
        extension = self._active[module]
        return (extension.show_help.im_func is not
                gaupol.Extension.show_help.im_func)

    def has_preferences_dialog_require(self, module):
        assert module in self._active
        assert module in self._metadata

    def has_preferences_dialog(self, module):
        """Return True if extension can display a preferences dialog."""
        extension = self._active[module]
        return (extension.show_preferences_dialog.im_func is not
                gaupol.Extension.show_preferences_dialog.im_func)

    def is_active(self, module):
        """Return ``True`` if extension is active, ``False`` if not."""
        return (module in self._active)

    def setup_extension_require(self, module, inferior=False):
        assert module in self._metadata

    def setup_extension(self, module, inferior=False):
        """Import and setup extension by module name.

        `inferior` should be ``True`` if called just to satisfy a dependency of
        another extension. Setup also all modules required by `module`. Return
        silently if `module` is already set up.
        """
        if module in self._active: return
        metadata = self._metadata[module]
        for dependency in metadata.get_field_list("Requires", ()):
            if not dependency in self._active:
                self.setup_extension(dependency, True)
            self._dependants[dependency].append(module)
        directory = os.path.dirname(metadata.path)
        sys.path.insert(0, directory)
        try: mobj = __import__(module, {}, {}, [])
        except ImportError:
            return traceback.print_exc()
        finally: sys.path.pop(0)
        for attribute in dir(mobj):
            if attribute.startswith("_"): continue
            value = getattr(mobj, attribute)
            if not inspect.isclass(value): continue
            if issubclass(value, gaupol.Extension):
                extension = value()
                extension.setup(self.application)
                self._active[module] = extension
                self._dependants[module] = []
        if inferior:
            self._inferior.append(module)
        else: # superior
            if module in self._inferior:
                self._inferior.remove(module)
        self.application.update_gui()

    def setup_extensions(self):
        """Import and setup all extensions configured as active."""
        for module in gaupol.conf.extensions.active:
            if not module in self._metadata:
                gaupol.conf.extensions.active.remove(module)
            else: self.setup_extension(module)

    def show_help_require(self, module):
        assert module in self._active
        assert module in self._metadata

    def show_help(self, module):
        """Show documentation on using extension."""
        extension = self._active[module]
        extension.show_help()

    def show_preferences_dialog_require(self, module, parent):
        assert module in self._active
        assert module in self._metadata

    def show_preferences_dialog(self, module, parent):
        """Show a preferences dialog for configuring extension."""
        extension = self._active[module]
        extension.show_preferences_dialog(parent)

    def teardown_extension_require(self, module, force=True):
        assert module in self._metadata

    def teardown_extension(self, module, force=True):
        """Teardown extension by module name.

        If not using `force` (which should only be used with care), raise
        :exc:`gaupol.DependencyError` if module is required by other modules.
        Teardown also no longer used dependencies of `module`. Return silently
        if `module` is already torn down.
        """
        if not module in self._active: return
        if self._dependants[module]:
            if not force:
                raise gaupol.DependencyError(
                    "Module %s is required by other modules"
                    % repr(module))

            for user in self._dependants[module]:
                self.teardown_extension(user)
        extension = self._active[module]
        extension.teardown(self.application)
        del self._active[module]
        del self._dependants[module]
        for dependency in self._dependants.keys():
            # Teardown no longer used dependencies.
            if module in self._dependants[dependency]:
                self._dependants[dependency].remove(module)
            if self._dependants[dependency]: continue
            if not dependency in self._inferior: continue
            self.teardown_extension(dependency)
        if module in self._inferior:
            self._inferior.remove(module)
        gaupol.conf.write_to_file()

    def teardown_extensions(self):
        """Teardown all active extensions."""
        for module in self._active.keys():
            self.teardown_extension(module, True)

    def update_extensions(self, page):
        """Call ``update`` on all active extensions."""
        for name, extension in self._active.iteritems():
            extension.update(self.application, page)


class ExtensionMetadata(aeidon.MetadataItem):

    """Extension metadata specified in a separate desktop-style file."""

    def __init__(self, fields=None):
        """Initialize an :class:`ExtensionMetadata` object."""
        aeidon.MetadataItem.__init__(self, fields)
        self.path = None



# Copyright (C) 2005-2009 Osmo Salomaa
#
# This file is part of Gaupol.
#
# Gaupol is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Gaupol is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Gaupol. If not, see <http://www.gnu.org/licenses/>.

import aeidon


class TestSubStationAlpha(aeidon.TestCase):

    text = ("All things weird are normal\n"
            "in this whore of cities.")

    def setup_method(self, method):
        self.markup = aeidon.tags.new(aeidon.formats.SSA)

    def test_bolden(self):
        assert self.markup.bolden(self.text) == (
            "{\\b1}All things weird are normal\n"
            "in this whore of cities.{\\b0}")

    def test_colorize(self):
        assert self.markup.colorize(self.text, "ccff00") == (
            "{\\c&H00ffcc&}All things weird are normal\n"
            "in this whore of cities.")

    def test_decode__bold(self):
        text = ("{\\b1}All things weird are normal\n"
                "in this whore of cities.{\\b0}")
        assert self.markup.decode(text) == (
            "<b>All things weird are normal\n"
            "in this whore of cities.</b>")

    def test_decode__color(self):
        text = ("All things weird are normal\n"
                "in {\\c&Hff&}this whore of cities.")
        assert self.markup.decode(text) == (
            "All things weird are normal\n"
            "in <color=#ff0000>this whore of cities.</color>")

    def test_decode__combined(self):
        text = ("{\\b1\\i1}All things weird are normal\n"
                "in this{\\i0\\b0} whore of cities.")
        assert self.markup.decode(text) == (
            "<b><i>All things weird are normal\n"
            "in this</i></b> whore of cities.")

    def test_decode__font(self):
        text = ("All things {\\fnSans}weird{\\r} are normal\n"
                "in this whore of cities.")
        assert self.markup.decode(text) == (
            "All things <font=Sans>weird</font> are normal\n"
            "in this whore of cities.")

    def test_decode__italic(self):
        text = ("{\\i1}All{\\r} things weird are normal\n"
                "in this whore of cities.")
        assert self.markup.decode(text) == (
            "<i>All</i> things weird are normal\n"
            "in this whore of cities.")

    def test_decode__reset(self):
        text = ("{\\b1\\i1}All{\\i0} things weird are normal\n"
                "{\\fs12}in this whore of cities{\\r}.")
        assert self.markup.decode(text) == (
            "<b><i>All</i> things weird are normal\n"
            "<size=12>in this whore of cities</size></b>.")

    def test_decode__size(self):
        text = ("All things weird are normal\n"
                "{\\fs12}in this whore of cities.")
        assert self.markup.decode(text) == (
            "All things weird are normal\n"
            "<size=12>in this whore of cities.</size>")

    def test_encode__bold(self):
        text = ("<b>All things weird are normal\n"
                "in this whore of cities.</b>")
        assert self.markup.encode(text) == (
            "{\\b1}All things weird are normal\n"
            "in this whore of cities.{\\b0}")

    def test_encode__color(self):
        text = ("<color=#ccccff>All things weird are normal\n"
                "in this whore of cities.</color>")
        assert self.markup.encode(text) == (
            "{\\c&Hffcccc&}All things weird are normal\n"
            "in this whore of cities.")

    def test_encode__font(self):
        text = ("<font=Sans>All things weird are normal\n"
                "in this whore of cities.</font>")
        assert self.markup.encode(text) == (
            "{\\fnSans}All things weird are normal\n"
            "in this whore of cities.")

    def test_encode__italic(self):
        text = ("All <i>things</i> weird are normal\n"
                "in <i>this</i> whore of cities.")
        assert self.markup.encode(text) == (
            "All {\\i1}things{\\i0} weird are normal\n"
            "in {\\i1}this{\\i0} whore of cities.")

    def test_encode__size(self):
        text = ("All things weird are normal\n"
                "in this whore of <size=12>cities</size>.")
        assert self.markup.encode(text) == (
            "All things weird are normal\n"
            "in this whore of {\\fs12}cities.")

    def test_encode__underline(self):
        text = ("All things weird are normal\n"
                "<u>in this whore of cities.</u>")
        assert self.markup.encode(text) == self.text

    def test_fontify(self):
        assert self.markup.fontify(self.text, "Sans") == (
            "{\\fnSans}All things weird are normal\n"
            "in this whore of cities.")

    def test_italic_tag(self):
        assert self.markup.italic_tag.match("{\\i1}")
        assert self.markup.italic_tag.match("{\\i0}")

    def test_italicize(self):
        assert self.markup.italicize(self.text, (0, 3)) == (
            "{\\i1}All{\\i0} things weird are normal\n"
            "in this whore of cities.")

    def test_scale(self):
        assert self.markup.scale(self.text, 12) == (
            "{\\fs12}All things weird are normal\n"
            "in this whore of cities.")

    def test_tag(self):
        assert self.markup.tag.match("{\\b500}")
        assert self.markup.tag.match("{\\c&H&}")



# Copyright (C) 2005-2008,2010 Osmo Salomaa
#
# This file is part of Gaupol.
#
# Gaupol is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Gaupol is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Gaupol. If not, see <http://www.gnu.org/licenses/>.

"""Cell renderer for time data in format ``[-]HH:MM:SS,SSS``."""

import gaupol
import gtk

__all__ = ("TimeCellRenderer",)


class TimeCellRenderer(gtk.CellRendererText):

    """Cell renderer for time data in format ``[-]HH:MM:SS.SSS``."""

    __gtype_name__ = "TimeCellRenderer"

    def __init__(self):
        """Initialize a :class:`TimeCellRenderer` object."""
        gtk.CellRendererText.__init__(self)
        self._in_editor_menu = False

    def _on_editor_focus_out_event(self, editor, *args):
        """End editing."""
        if self._in_editor_menu: return
        editor.remove_widget()
        self.emit("editing-canceled")

    def _on_editor_key_press_event(self, editor, event):
        """End editing if ``Enter`` or ``Escape`` pressed."""
        if event.state & (gtk.gdk.SHIFT_MASK | gtk.gdk.CONTROL_MASK): return
        if event.keyval in (gtk.keysyms.Return, gtk.keysyms.KP_Enter):
            editor.remove_widget()
            self.emit("edited", editor.get_data("path"), editor.get_text())
        if event.keyval == gtk.keysyms.Escape:
            editor.remove_widget()
            self.emit("editing-canceled")

    def _on_editor_populate_popup(self, editor, menu):
        """Disable "focus-out-event" ending editing."""
        self._in_editor_menu = True
        def on_menu_unmap(menu, self):
            self._in_editor_menu = False
        menu.connect("unmap", on_menu_unmap, self)

    def do_start_editing(self, event, widget, path, bg_area, cell_area, flags):
        """Initialize and a :class:`gaupol.TimeEntry` widget."""
        editor = gaupol.TimeEntry()
        editor.set_has_frame(False)
        editor.set_alignment(self.props.xalign)
        editor.modify_font(self.props.font_desc)
        editor.set_text(self.props.text)
        editor.select_region(0, -1)
        editor.set_data("path", path)
        editor.connect("focus-out-event", self._on_editor_focus_out_event)
        editor.connect("key-press-event", self._on_editor_key_press_event)
        editor.connect("populate-popup", self._on_editor_populate_popup)
        editor.show()
        return editor



# Copyright (C) 2005-2009 Osmo Salomaa
#
# This file is part of Gaupol.
#
# Gaupol is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Gaupol is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Gaupol. If not, see <http://www.gnu.org/licenses/>.

"""MicroDVD file."""

import aeidon
import re

__all__ = ("MicroDVD",)


class MicroDVD(aeidon.SubtitleFile):

    """MicroDVD file."""

    _re_line = re.compile(r"^\{(-?\d+)\}\{(-?\d+)\}(.*?)$")
    format = aeidon.formats.MICRODVD
    mode = aeidon.modes.FRAME

    def read(self):
        """Read file and return subtitles.

        Raise :exc:`IOError` if reading fails.
        Raise :exc:`UnicodeError` if decoding fails.
        """
        subtitles = []
        for line in self._read_lines():
            match = self._re_line.match(line)
            if match is not None:
                subtitle = self._get_subtitle()
                subtitle.start = int(match.group(1))
                subtitle.end = int(match.group(2))
                subtitle.main_text = match.group(3).replace("|", "\n")
                subtitles.append(subtitle)
            elif line.startswith("{DEFAULT}"):
                self.header = line
        return subtitles

    def write_to_file(self, subtitles, doc, fobj):
        """Write `subtitles` from `doc` to `fobj`.

        Raise :exc:`IOError` if writing fails.
        Raise :exc:`UnicodeError` if encoding fails.
        """
        n = self.newline.value
        if self.header.strip():
            fobj.write(self.header.replace("\n", n))
            fobj.write(n)
        for subtitle in subtitles:
            fobj.write("{%d}" % subtitle.start_frame)
            fobj.write("{%d}" % subtitle.end_frame)
            fobj.write(subtitle.get_text(doc).replace("\n", "|"))
            fobj.write(n)



# Copyright (C) 2007,2009 Osmo Salomaa
#
# This file is part of Gaupol.
#
# Gaupol is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Gaupol is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Gaupol. If not, see <http://www.gnu.org/licenses/>.

"""Website URLs.

:var BUG_REPORT_URL: Web page where to submit new bug reports
:var EXTENSIONS_URL: Web page listing third party extensions
:var HOMEPAGE_URL: Web page of the Gaupol project
:var PREVIEW_HELP_URL: Documentation on the preview function
:var REGEX_HELP_URL: Documentation on regular expressions
:var SPEECH_RECOGNITION_HELP_URL: Documentation on speech recognition
:var WIKI_URL: Wiki documentation
"""

import gaupol
import gtk

__all__ = ("BUG_REPORT_URL",
           "EXTENSIONS_URL",
           "HOMEPAGE_URL",
           "PREVIEW_HELP_URL",
           "REGEX_HELP_URL",
           "SPEECH_RECOGNITION_HELP_URL",
           "WIKI_URL")

BUG_REPORT_URL = "http://bugzilla.gnome.org/enter_bug.cgi?product=gaupol"
EXTENSIONS_URL = "http://live.gnome.org/Gaupol/Extensions"
HOMEPAGE_URL = "http://home.gna.org/gaupol/"
PREVIEW_HELP_URL = "http://live.gnome.org/Gaupol/Preview"
REGEX_HELP_URL = "http://docs.python.org/lib/re-syntax.html"
SPEECH_RECOGNITION_HELP_URL = "http://live.gnome.org/Gaupol/SpeechRecognition"
WIKI_URL = "http://live.gnome.org/Gaupol"

def _on_link_button_clicked(link_button, uri):
    return gaupol.util.show_uri(uri)
gtk.link_button_set_uri_hook(_on_link_button_clicked)



# Copyright (C) 2005-2008 Osmo Salomaa
#
# This file is part of Gaupol.
#
# Gaupol is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Gaupol is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Gaupol. If not, see <http://www.gnu.org/licenses/>.

import gaupol


class TestFormatAgent(gaupol.TestCase):

    def setup_method(self, method):
        self.application = self.new_application()
        page = self.application.get_current_page()
        page.view.set_focus(0, page.view.columns.MAIN_TEXT)
        page.view.select_rows((0, 1, 2))

    def test_on_toggle_dialogue_dashes_activate(self):
        self.application.get_action("toggle_dialogue_dashes").activate()

    def test_on_toggle_italicization_activate(self):
        self.application.get_action("toggle_italicization").activate()

    def test_on_use_lower_case_activate(self):
        self.application.get_action("use_lower_case").activate()

    def test_on_use_sentence_case_activate(self):
        self.application.get_action("use_sentence_case").activate()

    def test_on_use_title_case_activate(self):
        self.application.get_action("use_title_case").activate()

    def test_on_use_upper_case_activate(self):
        self.application.get_action("use_upper_case").activate()



# Copyright (C) 2006-2008 Osmo Salomaa
#
# This file is part of Gaupol.
#
# Gaupol is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Gaupol is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Gaupol. If not, see <http://www.gnu.org/licenses/>.

import aeidon


class PuppetMaster(object):

    def __init__(self):
        self.name = "master"

    def _invariant(self):
        pass


class TestDelegate(aeidon.TestCase):

    def setup_method(self, method):
        self.master = PuppetMaster()
        self.delegate = aeidon.Delegate(self.master)

    def test___getattr__(self):
        assert self.delegate.name == "master"

    def test___setattr____master(self):
        self.delegate.name = "slave"
        assert self.master.name == "slave"

    def test___setattr____delegate(self):
        self.delegate.none = None
        assert "none" in self.delegate.__dict__
        assert "none" not in self.master.__dict__

    def test__invariant(self):
        self.delegate._invariant()



# Copyright (C) 2011 Osmo Salomaa
#
# This file is part of Gaupol.
#
# Gaupol is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Gaupol is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Gaupol. If not, see <http://www.gnu.org/licenses/>.

import aeidon
import gaupol
import gtk
import os
import sys
import traceback


class TestAddBookmarkDialog(gaupol.TestCase):

    def run__dialog(self):
        self.dialog.run()
        self.dialog.destroy()

    def setup_method(self, method):
        directory = os.path.abspath(os.path.dirname(__file__))
        directory = os.path.abspath(os.path.join(directory, ".."))
        sys.path.insert(0, directory)
        try: mobj = __import__("bookmarks", {}, {}, [])
        except ImportError:
            return traceback.print_exc()
        finally: sys.path.pop(0)
        self.application = self.new_application()
        self.page = self.application.get_current_page()
        self.page.view.select_rows((1,))
        self.dialog = mobj.AddBookmarkDialog(gtk.Window(), self.page)
        self.dialog.show()

    def test_get_description(self):
        self.dialog._description_entry.set_text("test")
        text = self.dialog.get_description()
        assert text == "test"

    def test_get_row(self):
        self.dialog._subtitle_spin.set_value(3)
        row = self.dialog.get_row()
        assert row == 2


class TestBookmarksExtension(gaupol.TestCase):

    @aeidon.deco.monkey_patch(gaupol.util, "run_dialog")
    def bookmark_subtitles(self):
        gaupol.util.run_dialog = lambda *args: gtk.RESPONSE_OK
        action = self.application.get_action("add_bookmark")
        for i in range(5):
            self.page.view.select_rows((i,))
            action.activate()

    def import_side_pane(self):
        directory = os.path.abspath(os.path.dirname(__file__))
        directory = os.path.abspath(os.path.join(directory,
                                                 "..",
                                                 "..",
                                                 "side-pane"))
        sys.path.insert(0, directory)
        try: mobj = __import__("side-pane", {}, {}, [])
        except ImportError:
            return traceback.print_exc()
        finally: sys.path.pop(0)
        return(mobj)

    def setup_method(self, method):
        ## side-pane extension
        self.application = self.new_application()
        mobj_sp = self.import_side_pane()
        self.extension_sp = mobj_sp.SidePaneExtension()
        self.extension_sp.setup(self.application)
        ## bookmarks extension
        directory = os.path.abspath(os.path.dirname(__file__))
        directory = os.path.abspath(os.path.join(directory, ".."))
        sys.path.insert(0, directory)
        try: mobj = __import__("bookmarks", {}, {}, [])
        except ImportError:
            return traceback.print_exc()
        finally: sys.path.pop(0)
        self.extension = mobj.BookmarksExtension()
        self.extension.setup(self.application)
        self.page = self.application.get_current_page()
        self.page.view.select_rows((1,))

    def teardown_method(self, method):
        self.extension.teardown(self.application)
        self.extension_sp.teardown(self.application)
        gaupol.TestCase.teardown_method(self, self.application)

    @aeidon.deco.monkey_patch(gaupol.util, "run_dialog")
    def test__on_add_bookmark_activate(self):
        gaupol.util.run_dialog = lambda *args: gtk.RESPONSE_OK
        action = self.application.get_action("add_bookmark")
        action.activate()

    def test__on_application_page_added(self):
        self.application.open_main(self.new_subrip_file())

    def test__on_application_page_closed(self):
        self.application.close(self.page)

    def test__on_application_page_switched(self):
        self.application.open_main(self.new_subrip_file())
        for page in self.application.pages:
            self.application.set_current_page(page)

    def test__on_edit_bookmarks_activate(self):
        action = self.application.get_action("edit_bookmarks")
        action.activate()

    def test__on_next_bookmark_activate(self):
        self.bookmark_subtitles()
        action = self.application.get_action("next_bookmark")
        action.activate()
        action.activate()
        action.activate()

    def test__on_project_main_file_saved(self):
        self.application.save_main(self.page)

    def test__on_project_subtitles_inserted(self):
        self.bookmark_subtitles()
        self.page.project.insert_subtitles(range(7))

    def test__on_project_subtitles_removed(self):
        self.bookmark_subtitles()
        self.page.project.remove_subtitles(range(7))

    def test__on_previous_bookmark_activate(self):
        self.bookmark_subtitles()
        action = self.application.get_action("previous_bookmark")
        action.activate()
        action.activate()
        action.activate()

    def test__on_search_entry_changed(self):
        self.bookmark_subtitles()
        self.extension._search_entry.set_text("a")
        self.extension._search_entry.set_text("x")

    def test__on_tree_view_selection_changed(self):
        self.bookmark_subtitles()
        selection = self.extension._tree_view.get_selection()
        selection.select_path(1)
        selection.select_path(2)
        selection.select_path(3)

    def test__read_bookmarks(self):
        path = self.new_subrip_file()
        path_bookmarks = path.replace(".srt", ".gaupol-bookmarks")
        fobj = open(path, "a")
        fobj.write("1 test\n")
        fobj.write("2 test\n")
        fobj.write("3 test\n")
        fobj.close()
        self.application.open_main(path)

    def test_setup(self):
        page = self.application.side_pane.get_current_page()
        assert page is self.extension._side_container

    def test_teardown(self):
        self.extension.teardown(self.application)
        page = self.application.side_pane.get_current_page()
        assert page is None
        self.extension.setup(self.application)

    def test_update(self):
        self.extension.update(self.application, self.page)



# Copyright (C) 2006-2009 Osmo Salomaa
#
# This file is part of Gaupol.
#
# Gaupol is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Gaupol is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Gaupol. If not, see <http://www.gnu.org/licenses/>.

"""Baseclass and wrapper for :class:`gtk.Builder` constructed dialogs."""

import aeidon
import gtk
import os

__all__ = ("BuilderDialog",)


class BuilderDialog(object):

    """Baseclass and wrapper for `gtk.Builder` constructed dialogs.

    :cvar _widgets: List of names of widgets to be assigned as attributes

    All widgets defined in :attr:`_widgets` are assigned as instance variables
    with names preceded by a single underscore. All signals defined in the UI
    definition file are connected to ``self``. All :func:`getattr` calls not
    found in ``self`` are delegated to :attr:`self._dialog` allowing ``self``
    to look and act like a :class:`gtk.Dialog`.
    """

    _widgets = ()

    def __getattr__(self, name):
        """Return attribute from :attr:`_dialog`."""
        return getattr(self._dialog, name)

    def __init__(self, ui_file_path, connect_signals=True):
        """Initialize a :class:`BuilderDialog` object from `ui_file_path`."""
        if not os.path.isabs(ui_file_path):
            ui_file_path = os.path.join(aeidon.DATA_DIR, "ui", ui_file_path)
        self._builder = gtk.Builder()
        self._builder.set_translation_domain("gaupol")
        self._builder.add_from_file(ui_file_path)
        if connect_signals:
            self._builder.connect_signals(self)
        self._dialog = self._builder.get_object("dialog")
        self._set_attributes(self._widgets)

    def _set_attributes(self, widgets, prefix=None):
        """Assign all names in `widgets` as attributes of `self`."""
        for name in widgets:
            widget = self._builder.get_object(name)
            if (prefix is not None) and (name.startswith(prefix)):
                name = name.replace(prefix, "")
            setattr(self, "_%s" % name, widget)

    def run(self):
        """Show the dialog, run it and return response."""
        self._dialog.show()
        return self._dialog.run()



# Copyright (C) 2006-2007,2009 Osmo Salomaa
#
# This file is part of Gaupol.
#
# Gaupol is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Gaupol is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Gaupol. If not, see <http://www.gnu.org/licenses/>.

"""Data editing extension delegates of :class:`aeidon.Project`."""

from .clipboard import ClipboardAgent
from .edit      import EditAgent
from .format    import FormatAgent
from .open      import OpenAgent
from .position  import PositionAgent
from .preview   import PreviewAgent
from .register  import RegisterAgent
from .save      import SaveAgent
from .search    import SearchAgent
from .set       import SetAgent
from .text      import TextAgent
from .util      import UtilityAgent

__all__ = tuple(filter(lambda x: x.endswith("Agent"), dir()))



# Copyright (C) 2005-2009 Osmo Salomaa
#
# This file is part of Gaupol.
#
# Gaupol is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# Gaupol is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# Gaupol. If not, see <http://www.gnu.org/licenses/>.

"""Text markup for the MicroDVD format."""

import aeidon

__all__ = ("MicroDVD",)


class MicroDVD(aeidon.Markup):

    """Text markup for the MicroDVD format.

    The MicroDVD format contains a lot of markup tags, of which the following
    are of interest to us. Tags that start with a lower case letter end at the
    end of the line, upper case ones at the end of the subtitle.

     * ``{c:$BBGGRR}..., {C:$BBGGRR}...``
     * ``{f:NAME}......, {F:NAME}......``
     * ``{s:SIZE}......, {S:SIZE}......``
     * ``{y:STYLE}....., {Y:STYLE}.....``

    Note the reverse order of the color tag, ``BBGGRR`` instead of the normal
    ``RRGGBB``. ``STYLE`` in the ``y`` is a string containing one or more of
    the letters "b", "i" and "u" and any amount of any possible separators or
    other characters.
    """

    format = aeidon.formats.MICRODVD

    def _main_decode_ensure(self, value, text):
        assert self.tag.search(value) is None

    def _main_decode(self, text):
        """Return `text` with decodable markup decoded."""
        text = self._decode_b(text, r"\{([Yy]:b)\}(.*?)\{/\1\}", 2)
        text = self._decode_c(text, r"\{([Cc]:#(.*?))\}(.*?)\{/\1\}", 2, 3)
        text = self._decode_f(text, r"\{([Ff]:(.*?))\}(.*?)\{/\1\}", 2, 3)
        text = self._decode_i(text, r"\{([Yy]:i)\}(.*?)\{/\1\}", 2)
        text = self._decode_s(text, r"\{([Ss]:(.*?))\}(.*?)\{/\1\}", 2, 3)
        return self._decode_u(text, r"\{([Yy]:u)\}(.*?)\{/\1\}", 2)

    def _pre_decode(self, text):
        """Return `text` with markup prepared for decoding."""
        text = self._pre_decode_break(text)
        text = self._pre_decode_color(text)
        return self._pre_decode_close(text)

    def _pre_decode_break(self, text):
        """Return `text` with combined markup tags separated.

        For example, ``{y:biu}`` is replaced with ``{y:b}{y:i}{y:u}``.
        """
        pattern = r"\{([Yy]):([^}]{2,})\}"
        regex = self._get_regex(pattern)
        match = regex.search(text)
        if match is None: return text
        y = match.group(1)
        replacement = ""
        for m in ("b", "i", "u"):
            if m in match.group(2):
                replacement += "{%s:%s}" % (y, m)
        text = regex.sub(replacement, text, 1)
        return self._pre_decode_break(text)

    def _pre_decode_close(self, text):
        """Return `text` with all markup tags closed.

        The artificial closing tags are of form ``{/X:VALUE}``.
        """
        # Add lower case closing tags to the end of each line.
        lines = text.split("\n")
        re_tag = self._get_regex(r"\{([cfsy]:.*?)\}")
        for i, line in enumerate(lines):
            matches = [x for x in re_tag.finditer(line)]
            for j in reversed(range(len(matches))):
                lines[i] += "{/%s}" % matches[j].group(1)
        text = "\n".join(lines)
        # Add upper case closing tags to the end of the text.
        re_tag = self._get_regex(r"\{([CFSY]:.*?)\}")
        matches = [x for x in re_tag.finditer(text)]
        for j in reversed(range(len(matches))):
            text += "{/%s}" % matches[j].group(1)
        return text

    def _pre_decode_color_ensure(self, value, text):
        regex = self._get_regex(r"\{([Cc]:)\$([0-9A-Fa-f]{6})\}")
        assert regex.search(value) is None

    def _pre_decode_color(self, text):
        """Return `text` with colors converted to standard hexadecimal form.

        Color tags are converted from ``{c:$BBGGRR}`` to ``{c:#RRGGBB}``.
        """
        regex = self._get_regex(r"\{([Cc]:)\$([0-9A-Fa-f]{6})\}")
        match = regex.search(text)
        if match is None: return text
        color = match.group(2)
        color = "%s%s%s" % (color[4:], color[2:4], color[:2])
        text = regex.sub(r"{\1#%s}" % color, text, 1)
        return self._pre_decode_color(text)

    def _style(self, text, upper, lower, value, bounds=None):
        """Return `text` wrapped in ``upper`` or ``lower`` markup tag."""
        a, z = bounds or (0, len(text))
        prefix = text[:a].split("\n")[-1]
        suffix = text[z:].split("\n")[0]
        re_alpha = self._get_regex(r"\w")
        # Return plain text if bounds does not define an entire line or
        # subtitle and thus cannot be marked without side-effects.
        if re_alpha.search(prefix): return text
        if re_alpha.search(suffix): return text
        if (not "\n" in text) or ("\n" in text[a:z]):
            tag = "{%s:%s}" % (upper, value)
        else: # Single line marked in a multiline subtitle.
            tag = "{%s:%s}" % (lower, value)
        return "".join((text[:a], "%s%s" % (tag, text[a:])))

    def bolden(self, text, bounds=None):
        """Return bolded `text`."""
        return self._style(text, "Y", "y", "b", bounds)

    def colorize(self, text, color, bounds=None):
        """Return `text` colorized to hexadecimal value."""
        # Reverse the color value from RRGGBB to BBGGRR.
        color = "$%s%s%s" % (color[4:], color[2:4], color[:2])
        return self._style(text, "C", "c", color, bounds)

    def fontify(self, text, font, bounds=None):
        """Return `text` changed to `font`."""
        return self._style(text, "F", "f", font, bounds)

    @property
    def italic_tag(self):
        """Regular expression for an italic markup tag."""
        return self._get_regex(r"\{[Yy]:i\}")

    def italicize(self, text, bounds=None):
        """Return italicized `text`."""
        return self._style(text, "Y", "y", "i", bounds)

    def scale(self, text, size, bounds=None):
        """Return `text` scaled to `size`."""
        return self._style(text, "S", "s", str(size), bounds)

    @property
    def tag(self):
        """Regular expression for any markup tag."""
        return self._get_regex(r"\{[CFSYcfsy]:.*?\}")

    def underline(self, text, bounds=None):
        """Return underlined `text`."""
        return self._style(text, "Y", "y", "u", bounds)
